#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for initializing a mlonmcu environment."""

import os

from mlonmcu.environment.templates import get_template_names
from mlonmcu.environment.config import get_environments_dir, DEFAULTS
from mlonmcu.environment.init import initialize_environment
from .helper.parse import extract_config


def add_init_options(parser):
    init_parser = parser.add_argument_group("init options")
    init_parser.add_argument(
        "-n",
        "--name",
        metavar="NAME",
        nargs=1,
        type=str,
        default="",
        help="Environment name (default: %(default)s)",
    )
    init_parser.add_argument(
        "-t",
        "--template",
        metavar="TEMPLATE",
        nargs=1,
        type=str,
        # choices=get_template_names(),
        default=DEFAULTS["template"],
        help="Environment template name or path (default: %(default)s, available: "
        + ", ".join(get_template_names())
        + ")",
    )
    init_parser.add_argument(
        "DIR",
        nargs="?",
        type=str,
        default=get_environments_dir(),
        help="Environment directory (default: " + os.path.join(get_environments_dir(), "{NAME}") + ")",
    )
    init_parser.add_argument(
        "--non-interactive",
        dest="non_interactive",
        action="store_true",
        help="Do not ask questions interactively",
    )
    init_parser.add_argument(
        "--venv",
        default=None,
        action="store_true",
        help="Create virtual python environment",
    )
    init_parser.add_argument(
        "--clone-models",
        dest="clone_models",
        default=None,
        action="store_true",
        help="Clone models directory into environment",
    )
    init_parser.add_argument(
        "--register",
        default=None,
        action="store_true",
        help="Add environment to the list of environment for the current user",
    )
    init_parser.add_argument(
        "--allow-exists",
        dest="allow_exists",
        default=None,
        action="store_true",
        help="Allow overwriting an existing environment directory",
    )
    init_parser.add_argument(
        "-c",
        "--config",
        metavar="KEY=VALUE",
        nargs="+",
        action="append",
        help=(
            "Set a number of key-value pairs "
            "(do not put spaces before or after the = sign). "
            "If a value contains spaces, you should define "
            "it with double quotes: "
            'foo="this is a sentence". Note that '
            "values are always treated as strings."
        ),
    )


def get_parser(subparsers):
    """ "Define and return a subparser for the init subcommand."""
    parser = subparsers.add_parser("init", description="Initialize ML on MCU environment.")
    parser.set_defaults(func=handle)
    add_init_options(parser)
    return parser


def handle(args):
    """Callback function which will be called to process the init subcommand"""
    name = args.name[0] if isinstance(args.name, list) else args.name
    config, _ = extract_config(args)
    initialize_environment(
        args.DIR,
        name,
        create_venv=args.venv,
        interactive=not args.non_interactive,
        clone_models=args.clone_models,
        register=args.register,
        template=args.template[0] if isinstance(args.template, list) else args.template,
        allow_exists=args.allow_exists,
        config=config,
    )



"""Test used RISC-V GCC Toolchains"""

import pytest

# from mlonmcu.session.run import RunStage
from mlonmcu.testing.riscv_toolchain import (
    _get_target_config,
    _get_feature_vext_config,
    _get_feature_pext_config,
    _test_riscv_toolchain,
)

# Using muRISCV-NN feature to get a relatively good coverage
# It would be nice if we would have specific test models which utilize all relevant kernels
# TODO: also test Zvl* extensions

# MODELS = ["tinymlperf"]
MODELS = ["resnet"]


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", MODELS)
@pytest.mark.parametrize("target", ["myriscv_default"])
@pytest.mark.parametrize("xlen", [32, 64])
# @pytest.mark.parametrize("xlen", [32])
@pytest.mark.parametrize("fpu", ["none", "single", "double"])
@pytest.mark.parametrize("embedded,compressed,atomic,multiply", [(False, True, True, True)])
@pytest.mark.parametrize("feature_names", [[], ["muriscvnn"]])
def test_default(
    model_name, target, xlen, fpu, embedded, compressed, atomic, multiply, feature_names, user_context, models_dir
):
    config = {
        **_get_target_config(target, xlen, fpu, embedded, compressed, atomic, multiply),
    }

    _, artifacts = _test_riscv_toolchain(
        "mlif", "tflmi", target, user_context, model_name, models_dir, feature_names, config, "gcc"
    )


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", MODELS)
@pytest.mark.parametrize("target", ["myriscv_vector"])
# @pytest.mark.parametrize("vlen", [64, 128, 256, 512, 1024, 2048])
@pytest.mark.parametrize("vlen", [64, 128, 2048])
@pytest.mark.parametrize("elen", [32])
@pytest.mark.parametrize("spec", [1.0])
@pytest.mark.parametrize("fpu", ["none", "single", "double"])
@pytest.mark.parametrize("embedded,compressed,atomic,multiply", [(False, True, True, True)])
@pytest.mark.parametrize("feature_names", [["vext"], ["vext", "muriscvnn"]])
# @pytest.mark.parametrize("feature_names", [["vext"]])
def test_embedded_vector_32bit(
    model_name,
    target,
    vlen,
    elen,
    spec,
    fpu,
    embedded,
    compressed,
    atomic,
    multiply,
    feature_names,
    user_context,
    models_dir,
):
    if elen == 32:
        pytest.skip("ELEN 32 not supported due to compiler bug")

    config = {
        **_get_target_config(target, 32, fpu, embedded, compressed, atomic, multiply),
        **_get_feature_vext_config(vlen, elen, spec, True),
    }

    _, artifacts = _test_riscv_toolchain(
        "mlif", "tflmi", target, user_context, model_name, models_dir, feature_names, config, "gcc"
    )


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", MODELS)
@pytest.mark.parametrize("target", ["myriscv_vector"])
# @pytest.mark.parametrize("vlen", [64, 128, 256, 512, 1024, 2048])
@pytest.mark.parametrize("vlen", [64, 128, 2048])
@pytest.mark.parametrize("elen", [32, 64])
@pytest.mark.parametrize("spec", [1.0])
@pytest.mark.parametrize("fpu", ["none", "single", "double"])
@pytest.mark.parametrize("embedded,compressed,atomic,multiply", [(False, True, True, True)])
@pytest.mark.parametrize("feature_names", [["vext"], ["vext", "muriscvnn"]])
# @pytest.mark.parametrize("feature_names", [["vext"]])
def test_embedded_vector_64bit(
    model_name,
    target,
    vlen,
    elen,
    spec,
    fpu,
    embedded,
    compressed,
    atomic,
    multiply,
    feature_names,
    user_context,
    models_dir,
):
    if elen == 32:
        pytest.skip("ELEN 32 not supported due to compiler bug")
    if fpu == "single":
        pytest.skip("Single precision float not supported due to compiler bug")

    config = {
        **_get_target_config(target, 64, fpu, embedded, compressed, atomic, multiply),
        **_get_feature_vext_config(vlen, elen, spec, True),
    }

    _, artifacts = _test_riscv_toolchain(
        "mlif",
        "tflmi",
        target,
        user_context,
        model_name,
        models_dir,
        feature_names,
        config,
        "gcc",
    )


# @pytest.mark.slow
# @pytest.mark.user_context
# @pytest.mark.parametrize("model_name", MODELS)
# @pytest.mark.parametrize("target", ["myriscv_vector"])
# # @pytest.mark.parametrize("vlen", [64, 128, 256, 512, 1024, 2048])
# @pytest.mark.parametrize("vlen", [64, 128, 2048])
# @pytest.mark.parametrize("spec", [1.0])
# # @pytest.mark.parametrize("fpu", ["none", "single", "double"])
# @pytest.mark.parametrize("embedded,compressed,atomic,multiply", [(False, True, True, True)])
# @pytest.mark.parametrize("feature_names", [["vext"], ["vext", "muriscvnn"]])
# def test_vector_32bit(
#     model_name, target, vlen, spec, fpu, embedded, compressed, atomic,
#     multiply, feature_names, user_context, models_dir
# ):
#     config = {
#         **_get_target_config(target, 32, fpu, embedded, compressed, atomic, multiply),
#         **_get_feature_vext_config(vlen, 64, spec, False),
#         **_get_platform_config(),
#     }
#     _, artifacts = _test_riscv_toolchain(
#         "mlif", "tflmi", target, user_context, model_name, models_dir, feature_names, config, "gcc"
#     )


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", MODELS)
@pytest.mark.parametrize("target", ["myriscv_vector"])
# @pytest.mark.parametrize("vlen", [64, 128, 256, 512, 1024, 2048])
@pytest.mark.parametrize("vlen", [64, 128, 2048])
@pytest.mark.parametrize("spec", [1.0])
# @pytest.mark.parametrize("fpu", ["none", "single", "double"])
@pytest.mark.parametrize("fpu", ["double"])
@pytest.mark.parametrize("embedded,compressed,atomic,multiply", [(False, True, True, True)])
@pytest.mark.parametrize("feature_names", [["vext"], ["vext", "muriscvnn"]])
def test_vector_64bit(
    model_name, target, vlen, spec, fpu, embedded, compressed, atomic, multiply, feature_names, user_context, models_dir
):
    config = {
        **_get_target_config(target, 64, fpu, embedded, compressed, atomic, multiply),
        **_get_feature_vext_config(vlen, 64, spec, False),
    }
    _, artifacts = _test_riscv_toolchain(
        "mlif", "tflmi", target, user_context, model_name, models_dir, feature_names, config, "gcc"
    )


@pytest.mark.slow
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", MODELS)
@pytest.mark.parametrize("target", ["myriscv_packed"])
@pytest.mark.parametrize("xlen", [32, 64])
@pytest.mark.parametrize("spec", [0.96])
@pytest.mark.parametrize("fpu", ["none", "single", "double"])
@pytest.mark.parametrize("embedded,compressed,atomic,multiply", [(False, True, True, True)])
@pytest.mark.parametrize("feature_names", [["pext"], ["pext", "muriscvnn"]])
def test_packed(
    model_name, target, xlen, spec, fpu, embedded, compressed, atomic, multiply, feature_names, user_context, models_dir
):
    config = {
        **_get_target_config(target, xlen, fpu, embedded, compressed, atomic, multiply),
        **_get_feature_pext_config(spec),
    }
    _, artifacts = _test_riscv_toolchain(
        "mlif", "tflmi", target, user_context, model_name, models_dir, feature_names, config, "gcc"
    )


# def test_vector_packed():



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU ETISS/Pulpino Target definitions"""

import os
import re
import csv
from pathlib import Path

from mlonmcu.logging import get_logger
from mlonmcu.timeout import exec_timeout
from mlonmcu.config import str2bool, str2list, str2dict
from mlonmcu.artifact import Artifact, ArtifactFormat
from mlonmcu.feature.features import SUPPORTED_TVM_BACKENDS
from mlonmcu.setup.utils import execute
from mlonmcu.target.common import cli
from mlonmcu.target.metrics import Metrics
from mlonmcu.target.bench import add_bench_metrics
from .riscv import RISCVTarget

logger = get_logger()


class EtissTarget(RISCVTarget):
    """Target using a simple RISC-V VP running in the ETISS simulator"""

    FEATURES = RISCVTarget.FEATURES | {
        "gdbserver",
        "etissdbg",
        "trace",
        "log_instrs",
        "pext",
        "vext",
        "xcorev",
        "vanilla_accelerator",
    }

    DEFAULTS = {
        **RISCVTarget.DEFAULTS,
        "gdbserver_enable": False,
        "gdbserver_attach": False,
        "gdbserver_port": 2222,
        "debug_etiss": False,
        "trace_memory": False,
        # "plugins": ["PrintInstruction"],
        "plugins": [],
        "verbose": False,
        "cpu_arch": None,
        "rom_start": 0x0,
        "rom_size": 0x800000,  # 8 MB
        "ram_start": 0x800000,
        "ram_size": 0x4000000,  # 64 MB
        "cycle_time_ps": 31250,  # 32 MHz
        "enable_vext": False,
        "vext_spec": 1.0,
        "embedded_vext": False,
        "enable_pext": False,
        "pext_spec": 0.96,
        "vlen": 0,  # vectorization=off
        "elen": 32,
        "jit": None,
        "allow_error": False,
        "max_block_size": None,
        "enable_xcorevmac": False,
        "enable_xcorevmem": False,
        "enable_xcorevbi": False,
        "enable_xcorevalu": False,
        "enable_xcorevbitmanip": False,
        "enable_xcorevsimd": False,
        "enable_xcorevhwlp": False,
        "extra_int_config": {},
        "extra_bool_config": {},
        "extra_string_config": {},
        "extra_plugin_config": {},
    }
    REQUIRED = RISCVTarget.REQUIRED | {"etiss.src_dir", "etiss.install_dir", "etissvp.script"}

    def __init__(self, name="etiss", features=None, config=None):
        super().__init__(name, features=features, config=config)
        # TODO: make optional or move to mlonmcu pkg
        self.metrics_script = Path(self.etiss_src_dir) / "src" / "bare_etiss_processor" / "get_metrics.py"

    @property
    def etiss_src_dir(self):
        return self.config["etiss.src_dir"]

    @property
    def etiss_dir(self):
        return self.config["etiss.install_dir"]

    @property
    def etiss_script(self):
        return self.config["etissvp.script"]

    @property
    def gdbserver_enable(self):
        value = self.config["gdbserver_enable"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def gdbserver_attach(self):
        value = self.config["gdbserver_attach"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def gdbserver_port(self):
        return int(self.config["gdbserver_port"])

    @property
    def debug_etiss(self):
        value = self.config["debug_etiss"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def trace_memory(self):
        value = self.config["trace_memory"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def plugins(self):
        value = self.config["plugins"]
        return str2list(value) if isinstance(value, str) else value

    @property
    def verbose(self):
        value = self.config["verbose"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def rom_start(self):
        value = self.config["rom_start"]
        return int(value, 0) if not isinstance(value, int) else value

    @property
    def rom_size(self):
        value = self.config["rom_size"]
        return int(value, 0) if not isinstance(value, int) else value

    @property
    def ram_start(self):
        value = self.config["ram_start"]
        return int(value, 0) if not isinstance(value, int) else value

    @property
    def ram_size(self):
        value = self.config["ram_size"]
        return int(value, 0) if not isinstance(value, int) else value

    @property
    def cycle_time_ps(self):
        return int(self.config["cycle_time_ps"])

    @property
    def cpu_arch(self):
        if self.config.get("cpu_arch", None):
            return self.config["cpu_arch"]
        elif self.enable_pext or self.enable_vext:
            return "RV32IMACFDPV"
        else:
            return "RV32IMACFD"

    @property
    def enable_vext(self):
        value = self.config["enable_vext"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_pext(self):
        value = self.config["enable_pext"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevmac(self):
        value = self.config["enable_xcorevmac"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevmem(self):
        value = self.config["enable_xcorevmem"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevbi(self):
        value = self.config["enable_xcorevbi"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevalu(self):
        value = self.config["enable_xcorevalu"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevbitmanip(self):
        value = self.config["enable_xcorevbitmanip"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevsimd(self):
        value = self.config["enable_xcorevsimd"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_xcorevhwlp(self):
        value = self.config["enable_xcorevhwlp"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def vlen(self):
        return int(self.config["vlen"])

    @property
    def elen(self):
        return int(self.config["elen"])

    @property
    def jit(self):
        return self.config["jit"]

    @property
    def extra_bool_config(self):
        value = self.config["extra_bool_config"]
        return str2dict(value) if not isinstance(value, dict) else value

    @property
    def extra_int_config(self):
        value = self.config["extra_int_config"]
        return str2dict(value) if not isinstance(value, dict) else value

    @property
    def extra_string_config(self):
        value = self.config["extra_string_config"]
        return str2dict(value) if not isinstance(value, dict) else value

    @property
    def extra_plugin_config(self):
        value = self.config["extra_plugin_config"]
        return str2dict(value) if not isinstance(value, dict) else value

    @property
    def extensions(self):
        exts = super().extensions
        required = set()
        if "xcorev" not in exts:
            if self.enable_xcorevmac:
                required.add("xcvmac")
            if self.enable_xcorevmem:
                required.add("xcvmem")
            if self.enable_xcorevbi:
                required.add("xcvbi")
            if self.enable_xcorevalu:
                required.add("xcvalu")
            if self.enable_xcorevbitmanip:
                required.add("xcvbitmanip")
            if self.enable_xcorevsimd:
                required.add("xcvsimd")
            if self.enable_xcorevhwlp:
                required.add("xcvhwlp")
        for ext in required:
            if ext not in exts:
                exts.add(ext)
        return exts

    @property
    def attr(self):
        attrs = super().attr.split(",")
        # attrs.append("+unaligned-scalar-mem")
        # attrs = [x for x in attrs if x != "+c"]
        # attrs.append("-c")
        if self.enable_xcorevmac:
            if "xcorevmac" not in attrs:
                attrs.append("+xcvmac")
        if self.enable_xcorevmem:
            if "xcorevmem" not in attrs:
                attrs.append("+xcvmem")
        if self.enable_xcorevbi:
            if "xcorevbi" not in attrs:
                attrs.append("+xcvbi")
        if self.enable_xcorevalu:
            if "xcorevalu" not in attrs:
                attrs.append("+xcvalu")
        if self.enable_xcorevbitmanip:
            if "xcorevbitmanip" not in attrs:
                attrs.append("+xcvbitmanip")
        if self.enable_xcorevsimd:
            if "xcorevsimd" not in attrs:
                attrs.append("+xcvsimd")
        if self.enable_xcorevhwlp:
            if "xcorevhwlp" not in attrs:
                attrs.append("+xcvhwlp")
        if self.enable_vext and f"+zvl{self.vlen}b" not in attrs:
            attrs.append(f"+zvl{self.vlen}b")
        return ",".join(attrs)

    @property
    def allow_error(self):
        value = self.config["allow_error"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def vext_spec(self):
        return float(self.config["vext_spec"])

    @property
    def embedded_vext(self):
        value = self.config["embedded_vext"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def pext_spec(self):
        return float(self.config["pext_spec"])

    @property
    def max_block_size(self):
        value = self.config["max_block_size"]
        if isinstance(value, str):
            value = int(value)
        return value

    def get_ini_bool_config(self):
        ret = {
            "arch.enable_semihosting": True,
        }
        ret.update(self.extra_string_config)
        return ret

    def get_ini_string_config(self):
        ret = {
            "arch.cpu": self.cpu_arch,
        }
        if self.jit is not None:
            ret["jit.type"] = f"{self.jit}JIT"
        ret.update(self.extra_string_config)
        return ret

    def get_ini_int_config(self):
        ret = {
            "simple_mem_system.memseg_origin_00": self.rom_start,
            "simple_mem_system.memseg_length_00": self.rom_size,
            "simple_mem_system.memseg_origin_01": self.ram_start,
            "simple_mem_system.memseg_length_01": self.ram_size,
            "arch.cpu_cycle_time_ps": self.cycle_time_ps,
        }
        if self.max_block_size:
            ret["etiss.max_block_size"] = self.max_block_size
        if self.has_fpu:
            # TODO: do not hardcode cpu_arch
            # TODO: i.e. use cpu_arch_lower
            ret["arch.rv32imacfdpv.mstatus_fs"] = 1
        if self.enable_vext:
            ret["arch.rv32imacfdpv.mstatus_vs"] = 1
            if self.vlen > 0:
                ret["arch.rv32imacfdpv.vlen"] = self.vlen
            if self.elen > 0:
                ret["arch.rv32imacfdpv.elen"] = self.elen
        ret.update(self.extra_int_config)
        return ret

    def get_ini_plugin_config(self):
        ret = {}
        if self.gdbserver_enable:
            # This could also be accomplished using `--plugin.gdbserver.port` on the cmdline
            ret["gdbserver"] = {
                "port": self.gdbserver_port,
            }
        ret.update(self.extra_plugin_config)  # TODO: merge nested dict instead of overriding
        return ret

    def write_ini(self, path):
        # TODO: Either create artifact for ini or prefer to use cmdline args.
        with open(path, "w") as f:
            ini_bool = self.get_ini_bool_config()
            if len(ini_bool) > 0:
                f.write("[BoolConfigurations]\n")
                for key, value in ini_bool.items():
                    assert isinstance(value, bool)
                    val = "true" if value else "false"
                    f.write(f"{key}={val}\n")
            ini_string = self.get_ini_string_config()
            if len(ini_string) > 0:
                f.write("[StringConfigurations]\n")
                for key, value in ini_string.items():
                    assert isinstance(value, str)
                    f.write(f"{key}={value}\n")
            ini_int = self.get_ini_int_config()
            if len(ini_int) > 0:
                f.write("[IntConfigurations]\n")
                for key, value in ini_int.items():
                    assert isinstance(value, int)
                    f.write(f"{key}={value}\n")
            ini_plugin = self.get_ini_plugin_config()
            if len(ini_plugin) > 0:
                for name, cfg in ini_plugin.items():
                    f.write(f"[Plugin {name}]\n")
                    for key, value in cfg.items():
                        if isinstance(value, bool):
                            val = "true" if value else "false"
                        else:
                            val = value
                        f.write(f"plugin.{name}.{key}={val}\n")

    def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
        """Use target to execute a executable with given arguments"""
        etiss_script_args = []
        if len(self.extra_args) > 0:
            etiss_script_args.extend(self.extra_args.split(" "))

        # TODO: this is outdated
        # TODO: validate features (attach xor noattach!)
        if self.debug_etiss:
            etiss_script_args.append("gdb")
        if self.gdbserver_enable:
            etiss_script_args.append("tgdb")
            if not self.gdbserver_attach:
                etiss_script_args.append("noattach")
        if self.trace_memory:
            etiss_script_args.append("trace")
            etiss_script_args.append("nodmi")
        if self.verbose:
            etiss_script_args.append("v")
        # Alternative to stdout parsing: etiss_script_args.append("--vp.stats_file_path=stats.json")

        # TODO: working directory?
        etiss_ini = os.path.join(cwd, "custom.ini")
        self.write_ini(etiss_ini)
        etiss_script_args.append("-i" + etiss_ini)
        for plugin in self.plugins:
            etiss_script_args.extend(["-p", plugin])

        # if self.timeout_sec > 0:
        if False:
            ret = exec_timeout(
                self.timeout_sec,
                execute,
                Path(self.etiss_script).resolve(),
                program,
                *etiss_script_args,
                *args,
                cwd=cwd,
                **kwargs,
            )
        else:
            ret = execute(
                Path(self.etiss_script).resolve(),
                program,
                *etiss_script_args,
                *args,
                cwd=cwd,
                **kwargs,
            )
        return ret, []

    def parse_exit(self, out):
        exit_code = super().parse_exit(out)
        if exit_code is None:
            # legacy
            exit_match = re.search(r"exit called with code: (.*)", out)
            if exit_match:
                exit_code = int(exit_match.group(1))
        return exit_code

    def parse_stdout(self, out, metrics, exit_code=0):
        add_bench_metrics(out, metrics, exit_code != 0, target_name=self.name)
        error_match = re.search(r"ETISS: Error: (.*)", out)
        if error_match:
            error_msg = error_match.group(1)
            if self.allow_error:
                logger.error(f"An ETISS Error occured during simulation: {error_msg}")
            else:
                raise RuntimeError(f"An ETISS Error occured during simulation: {error_msg}")
        sim_insns = re.search(r"CPU Cycles \(estimated\): (.*)", out)
        sim_insns = int(float(sim_insns.group(1)))
        metrics.add("Simulated Instructions", sim_insns, True)
        mips = None  # TODO: parse mips?
        mips_match = re.search(r"MIPS \(estimated\): (.*)", out)
        if mips_match:
            mips_str = mips_match.group(1)
            mips = float(mips_str)
        if mips:
            metrics.add("MIPS", mips, optional=True)

    def get_metrics(self, elf, directory, *args, handle_exit=None):
        out = ""
        if self.trace_memory:
            trace_file = os.path.join(directory, "dBusAccess.csv")
            if os.path.exists(trace_file):
                os.remove(trace_file)
        else:
            trace_file = None

        metrics_file = os.path.join(directory, "metrics.csv")
        if os.path.exists(metrics_file):
            os.remove(metrics_file)

        def _handle_exit(code, out=None):
            assert out is not None
            temp = self.parse_exit(out)
            # TODO: before or after?
            if temp is None:
                temp = code
            if handle_exit is not None:
                temp = handle_exit(temp, out=out)
            return temp

        artifacts = []

        if self.print_outputs:
            out_, artifacts_ = self.exec(elf, *args, cwd=directory, live=True, handle_exit=_handle_exit)
            out += out_
            artifacts += artifacts_
        else:
            out_, artifacts_ = self.exec(
                elf, *args, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=_handle_exit
            )
            out += out_
            artifacts += artifacts_
        # TODO: get exit code
        exit_code = 0
        metrics = Metrics()
        self.parse_stdout(out, metrics, exit_code=exit_code)

        get_metrics_args = [elf]
        etiss_ini = os.path.join(directory, "custom.ini")
        if os.path.exists(etiss_ini):
            get_metrics_args.extend(["--ini", etiss_ini])
        if trace_file:
            get_metrics_args.extend(["--trace", trace_file])
        get_metrics_args.extend(["--out", metrics_file])
        if self.print_outputs:
            out += execute(self.metrics_script.resolve(), *get_metrics_args, live=True)
        else:
            out += execute(
                self.metrics_script.resolve(),
                *get_metrics_args,
                live=False,
                cwd=directory,
                print_func=lambda *args, **kwargs: None,
            )

        metrics_file = os.path.join(directory, "metrics.csv")
        with open(metrics_file, "r") as handle:
            metrics_content = handle.read()
            lines = metrics_content.splitlines()
            reader = csv.DictReader(lines)
            data = list(reader)[0]

            def get_rom_sizes(data):
                assert "rom_rodata" in data
                rom_ro = int(data["rom_rodata"])
                assert "rom_code" in data
                rom_code = int(data["rom_code"])
                assert "rom_misc" in data
                rom_misc = int(data["rom_misc"])

                rom_total = rom_ro + rom_code + rom_misc
                return rom_total, rom_ro, rom_code, rom_misc

            def get_ram_sizes(data):
                assert "ram_data" in data
                ram_data = int(data["ram_data"])
                assert "ram_zdata" in data
                ram_zdata = int(data["ram_zdata"])
                ram_total = ram_data + ram_zdata
                if self.trace_memory:
                    assert "ram_stack" in data
                    ram_stack = int(data["ram_stack"])
                    assert "ram_heap" in data
                    ram_heap = int(data["ram_heap"])
                    ram_total += ram_stack + ram_heap
                else:
                    ram_stack = None
                    ram_heap = None
                return ram_total, ram_data, ram_zdata, ram_stack, ram_heap

            rom_total, rom_ro, rom_code, rom_misc = get_rom_sizes(data)
            ram_total, ram_data, ram_zdata, ram_stack, ram_heap = get_ram_sizes(data)
            metrics.add("Total ROM", rom_total)
            metrics.add("Total RAM", ram_total)
            metrics.add("ROM read-only", rom_ro)
            metrics.add("ROM code", rom_code)
            metrics.add("ROM misc", rom_misc)
            metrics.add("RAM data", ram_data)
            metrics.add("RAM zero-init data", ram_zdata)
            if self.trace_memory:
                metrics.add("RAM stack", ram_stack)
                metrics.add("RAM heap", ram_heap)

        ini_content = open(etiss_ini, "r").read()
        ini_artifact = Artifact("custom.ini", content=ini_content, fmt=ArtifactFormat.TEXT)
        artifacts.append(ini_artifact)

        return metrics, out, artifacts

    def get_target_system(self):
        return self.name

    def get_platform_defs(self, platform):
        assert platform == "mlif"
        ret = super().get_platform_defs(platform)
        ret["MEM_ROM_ORIGIN"] = self.rom_start
        ret["MEM_ROM_LENGTH"] = self.rom_size
        ret["MEM_RAM_ORIGIN"] = self.ram_start
        ret["MEM_RAM_LENGTH"] = self.ram_size
        if self.enable_pext:
            major, minor = str(self.pext_spec).split(".")[:2]
            ret["RISCV_RVP_MAJOR"] = major
            ret["RISCV_RVP_MINOR"] = minor
        if self.enable_vext:
            major, minor = str(self.vext_spec).split(".")[:2]
            ret["RISCV_RVV_MAJOR"] = major
            ret["RISCV_RVV_MINOR"] = minor
            ret["RISCV_RVV_VLEN"] = self.vlen
        return ret

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = super().get_backend_config(
            backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
        )
        if backend in SUPPORTED_TVM_BACKENDS:
            if optimized_layouts:
                if self.enable_pext or self.enable_vext:
                    ret.update(
                        {
                            # Warning: passing kernel layouts does not work with upstream TVM
                            # TODO: allow passing map?
                            "desired_layout": "NHWC:HWOI",
                        }
                    )
        return ret


if __name__ == "__main__":
    cli(target=EtissTarget)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

pytest_plugins = [
    "tests.fixtures",
]


def pytest_addoption(parser):
    parser.addoption("--run-slow", action="store_true", default=False, help="run slow tests")
    parser.addoption("--run-user-context", action="store_true", default=False, help="run tests using user context")
    parser.addoption("--run-hardware", action="store_true", default=False, help="run tests using real hardware")


# def pytest_configure(config):
#     config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    slow = False
    if config.getoption("--run-slow"):
        # --run-slow given in cli: do not skip slow tests
        slow = True
    user_context = False
    if config.getoption("--run-user-context"):
        # --run-context given in cli: do not skip user context tests
        user_context = True
    hardware = False
    if config.getoption("--run-hardware"):
        # --run-hardware given in cli: do not skip hardware tests
        hardware = True
    skip_slow = pytest.mark.skip(reason="need --run-slow option to run")
    skip_user_context = pytest.mark.skip(reason="need --run-context option to run")
    skip_hardware = pytest.mark.skip(reason="need --run-hardware option to run")
    for item in items:
        if not slow and "slow" in item.keywords:
            item.add_marker(skip_slow)
        if not user_context and "user_context" in item.keywords:
            item.add_marker(skip_user_context)
        if not hardware and "hardware" in item.keywords:
            item.add_marker(skip_hardware)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""TVM Build Platform"""
from ..platform import BuildPlatform
from .tvm_backend import create_tvm_platform_backend, get_tvm_platform_backends


class TvmBuildPlatform(BuildPlatform):
    """TVM build platform class."""

    def create_backend(self, name):
        supported = self.get_supported_backends()
        assert name in supported, f"{name} is not a valid TVM platform backend"
        base = supported[name]
        return create_tvm_platform_backend(name, self, base=base)

    def get_supported_backends(self):
        backend_names = get_tvm_platform_backends()
        return backend_names



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU GVSOC/Pulp or Pulpissimo Target definitions"""

import os
import re
from pathlib import Path

from mlonmcu.logging import get_logger
from mlonmcu.feature.features import SUPPORTED_TVM_BACKENDS
from mlonmcu.setup.utils import execute
from mlonmcu.target.common import cli
from mlonmcu.target.metrics import Metrics
from .riscv import RISCVTarget
from .util import update_extensions_pulp
import shutil

logger = get_logger()


class GvsocPulpTarget(RISCVTarget):
    """Target using a Pulpino-like VP running in the GVSOC simulator"""

    FEATURES = RISCVTarget.FEATURES | {"log_instrs", "xpulp"}

    DEFAULTS = {
        **RISCVTarget.DEFAULTS,
        "abi": "ilp32",
        "extensions": ["i", "m", "c"],  # TODO overwrite extensions elegantly
        "fpu": None,
        "xpulp_version": None,  # None means that xpulp extension is not used
        "model": "pulp",
    }

    REQUIRED = RISCVTarget.PUPL_GCC_TOOLCHAIN_REQUIRED | {
        "gvsoc.exe",
        "pulp_freertos.support_dir",
        "pulp_freertos.config_dir",
        "pulp_freertos.install_dir",
    }

    def __init__(self, name="gvsoc_pulp", features=None, config=None):
        super().__init__(name, features=features, config=config)

    @property
    def gvsoc_script(self):
        return Path(self.config["gvsoc.exe"])

    @property
    def gvsoc_folder(self):
        return Path(self.config["gvsoc.exe"]).parent / "gvsoc"

    @property
    def pulp_freertos_support_dir(self):
        return Path(self.config["pulp_freertos.support_dir"])

    @property
    def pulp_freertos_config_dir(self):
        return Path(self.config["pulp_freertos.config_dir"])

    @property
    def pulp_freertos_install_dir(self):
        return Path(self.config["pulp_freertos.install_dir"])

    @property
    def model(self):
        assert self.config["model"] in ["pulp", "pulpissimo"]
        return self.config["model"]

    @property
    def xpulp_version(self):
        value = self.config["xpulp_version"]
        return value

    @property
    def abi(self):
        value = self.config["abi"]
        return value

    @property
    def extensions(self):
        exts = super().extensions
        return update_extensions_pulp(exts, xpulp_version=self.xpulp_version)

    def gvsoc_preparation_env(self):
        return {
            "PULP_RISCV_GCC_TOOLCHAIN": str(self.pulp_gcc_prefix),
            "PULP_CURRENT_CONFIG": f"{self.model}@config_file=chips/{self.model}/{self.model}.json",
            "PULP_CONFIGS_PATH": str(self.pulp_freertos_config_dir),
            "PYTHONPATH": str(self.pulp_freertos_install_dir / "python"),
            "INSTALL_DIR": str(self.pulp_freertos_install_dir),
            "ARCHI_DIR": str(self.pulp_freertos_support_dir / "archi" / "include"),
            "SUPPORT_ROOT": str(self.pulp_freertos_support_dir),
        }

    def get_basic_gvsoc_simulating_arg(self, program):
        gvsoc_simulating_arg = []
        gvsoc_simulating_arg.append(f"--dir={program.parent / 'gvsim'}")
        gvsoc_simulating_arg.append(f"--config-file={self.model}@config_file=chips/{self.model}/{self.model}.json")
        gvsoc_simulating_arg.append("--platform=gvsoc")
        gvsoc_simulating_arg.append(f"--binary={program.stem}")
        gvsoc_simulating_arg.append("prepare")
        gvsoc_simulating_arg.append("run")
        return gvsoc_simulating_arg

    def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
        """Use target to execute an executable with given arguments"""
        # create a new folder and copy the compiled program into it
        gvsimDir = program.parent / "gvsim"
        if not os.path.exists(gvsimDir):
            os.makedirs(gvsimDir)
        shutil.copyfile(program, gvsimDir / program.stem)

        # prepare simulation by compile gvsoc according to the chosen chip
        gvsoc_compile_args = []
        gvsoc_compile_args.append("build")
        gvsoc_compile_args.append(f"ARCHI_DIR={self.pulp_freertos_support_dir / 'archi' / 'include'}")

        env = os.environ.copy()
        env.update(self.gvsoc_preparation_env())

        gvsoc_compile_retval = execute(
            "make",
            *gvsoc_compile_args,
            env=env,
            cwd=self.gvsoc_folder,
            *args,
            **kwargs,
        )

        # run simulation
        gvsoc_simulating_arg = self.get_basic_gvsoc_simulating_arg(program)
        if len(self.extra_args) > 0:
            if isinstance(self.extra_args, str):
                extra_args = self.extra_args.split(" ")
            else:
                extra_args = self.extra_args
            gvsoc_simulating_arg.extend(extra_args)

        env = os.environ.copy()
        env.update({"PULP_RISCV_GCC_TOOLCHAIN": str(self.pulp_gcc_prefix)})
        simulation_retval = execute(
            str(self.gvsoc_script),
            *gvsoc_simulating_arg,
            env=env,
            cwd=cwd,
            *args,
            **kwargs,
        )
        return gvsoc_compile_retval + simulation_retval

    def parse_stdout(self, out):
        cpu_cycles = re.search(r"Total Cycles: (.*)", out)
        if not cpu_cycles:
            logger.warning("unexpected script output (cycles)")
            cycles = None
        else:
            cycles = int(float(cpu_cycles.group(1)))

        cpu_instructions = re.search(r"Total Instructions: (.*)", out)
        if not cpu_instructions:
            logger.warning("unexpected script output (instructions)")
            cpu_instructions = None
        else:
            cpu_instructions = int(float(cpu_instructions.group(1)))
        return cycles, cpu_instructions

    def get_metrics(self, elf, directory, *args, handle_exit=None):
        out = ""
        if self.print_outputs:
            out += self.exec(elf, *args, cwd=directory, live=True, handle_exit=handle_exit)
        else:
            out += self.exec(
                elf, *args, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=handle_exit
            )
        cycles, instructions = self.parse_stdout(out)
        metrics = Metrics()
        metrics.add("Cycles", cycles)
        metrics.add("Instructions", instructions)
        return metrics, out, []

    def get_target_system(self):
        return f"gvsoc_{self.model}"

    def get_platform_defs(self, platform):
        assert platform == "mlif"
        ret = super().get_platform_defs(platform)
        # ret["RISCV_ARCH"] = "rv32imcxpulpv3"
        ret["RISCV_ABI"] = self.abi
        return ret

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = super().get_backend_config(
            backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
        )
        if backend in SUPPORTED_TVM_BACKENDS:
            ret.update({"target_model": f"{self.model}-{self.arch}"})
        return ret


if __name__ == "__main__":
    cli(target=GvsocPulpTarget)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_mlif(context: MlonMcuContext, params=None):
    return context.environment.has_platform("mlif")


@Tasks.provides(["mlif.src_dir"])
@Tasks.validate(_validate_mlif)
@Tasks.register(category=TaskType.PLATFORM)
def clone_mlif(context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()):
    """Clone the MLonMCU SW repository."""
    mlifName = utils.makeDirName("mlif")
    mlifSrcDir = context.environment.paths["deps"].path / "src" / mlifName
    if rebuild or not utils.is_populated(mlifSrcDir):
        mlifRepo = context.environment.repos["mlif"]
        utils.clone_wrapper(mlifRepo, mlifSrcDir, refresh=rebuild)
    context.cache["mlif.src_dir"] = mlifSrcDir



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import time
import tempfile
import multiprocessing

# from abc import ABC
from abc import abstractmethod
from pathlib import Path
from filelock import FileLock
from typing import Tuple, List

from mlonmcu.config import filter_config
from mlonmcu.feature.features import get_matching_features
from mlonmcu.feature.type import FeatureType
from mlonmcu.target.metrics import Metrics
from mlonmcu.target.elf import get_results as get_static_mem_usage
from mlonmcu.logging import get_logger
from mlonmcu.config import str2bool
from mlonmcu.artifact import Artifact, ArtifactFormat

logger = get_logger()


class Platform:
    """Abstract platform class."""

    FEATURES = set()

    DEFAULTS = {
        "print_outputs": False,
    }

    REQUIRED = set()
    OPTIONAL = set()

    def __init__(self, name, features=None, config=None):
        self.name = name
        self.config = config if config else {}
        self.definitions = {}
        self.features = self.process_features(features)
        self.config = filter_config(self.config, self.name, self.DEFAULTS, self.OPTIONAL, self.REQUIRED)
        self.artifacts = []

    def init_directory(self, path=None, context=None):
        raise NotImplementedError

    @property
    def supports_build(self):
        return False

    @property
    def supports_tune(self):
        return False

    @property
    def supports_compile(self):
        return False

    @property
    def supports_flash(self):
        return False

    @property
    def supports_monitor(self):
        return False

    @property
    def print_outputs(self):
        value = self.config["print_outputs"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def process_features(self, features):
        if features is None:
            return []
        features = get_matching_features(features, FeatureType.PLATFORM)
        for feature in features:
            # assert feature.name in self.FEATURES, f"Incompatible feature: {feature.name}"
            if feature.name in self.FEATURES:
                feature.used = True
                feature.add_platform_config(self.name, self.config)
                feature.add_platform_defs(self.name, self.definitions)
        return features

    def get_supported_backends(self):
        return []

    def get_supported_targets(self):
        return []


class BuildPlatform(Platform):
    """Abstract build platform class."""

    @property
    def supports_build(self):
        return True

    def export_artifacts(self, path):
        assert len(self.artifacts) > 0, "No artifacts found, please run generate_artifacts() first"

        if not isinstance(path, Path):
            path = Path(path)
        assert (
            path.is_dir()
        ), "The supplied path does not exists."  # Make sure it actually exists (we do not create it by default)
        for artifact in self.artifacts:
            artifact.export(path)


class TunePlatform(Platform):
    """Abstract tune platform class."""

    @property
    def supports_tune(self):
        return True

    def export_artifacts(self, path):
        assert len(self.artifacts) > 0, "No artifacts found, please run generate_artifacts() first"

        if not isinstance(path, Path):
            path = Path(path)
        assert (
            path.is_dir()
        ), "The supplied path does not exists."  # Make sure it actually exists (we do not create it by default)
        for artifact in self.artifacts:
            artifact.export(path)

    @abstractmethod
    def _tune_model(self, model_path, backend, target):
        raise NotImplementedError

    def tune_model(self, model_path, backend, target):
        start_time = time.time()
        artifacts, metrics = self._tune_model(model_path, backend, target)
        # TODO: do something with out?
        end_time = time.time()
        diff = end_time - start_time
        if len(metrics) == 0:
            metrics = {"default": Metrics()}
        for name, metrics_ in metrics.items():
            if name == "default":
                metrics_.add("Tune Stage Time [s]", diff, True)
            content = metrics_.to_csv(include_optional=True)  # TODO: store df instead?
            artifact = Artifact("tune_metrics.csv", content=content, fmt=ArtifactFormat.TEXT, flags=["metrics"])
            if name not in artifacts:
                artifacts[name] = []
            artifacts[name].append(artifact)
        return artifacts


class CompilePlatform(Platform):
    """Abstract compile platform class."""

    FEATURES = Platform.FEATURES | {"debug"}

    DEFAULTS = {
        **Platform.DEFAULTS,
        "debug": False,
        "build_dir": None,
        "num_threads": multiprocessing.cpu_count(),
    }

    @property
    def supports_compile(self):
        return True

    @property
    def debug(self):
        value = self.config["debug"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def num_threads(self):
        return max(1, int(self.config["num_threads"]))

    def get_metrics(self, elf):
        static_mem = get_static_mem_usage(elf)
        rom_ro, rom_code, rom_misc, ram_data, ram_zdata = (
            static_mem["rom_rodata"],
            static_mem["rom_code"],
            static_mem["rom_misc"],
            static_mem["ram_data"],
            static_mem["ram_zdata"],
        )
        rom_total = rom_ro + rom_code + rom_misc
        ram_total = ram_data + ram_zdata
        metrics = Metrics()
        metrics.add("Total ROM", rom_total)
        metrics.add("Total RAM", ram_total)
        metrics.add("ROM read-only", rom_ro)
        metrics.add("ROM code", rom_code)
        metrics.add("ROM misc", rom_misc)
        metrics.add("RAM data", ram_data)
        metrics.add("RAM zero-init data", ram_zdata)
        return metrics

    @abstractmethod
    def generate(self, src, target, model=None) -> Tuple[dict, dict]:
        raise NotImplementedError

    def generate_artifacts(self, src, target, model=None) -> List[Artifact]:
        start_time = time.time()
        artifacts, metrics = self.generate(src, target, model=None)
        # TODO: do something with out?
        end_time = time.time()
        diff = end_time - start_time
        if len(metrics) == 0:
            metrics = {"default": Metrics()}
        for name, metrics_ in metrics.items():
            if name == "default":
                metrics_.add("Compile Stage Time [s]", diff, True)
            content = metrics_.to_csv(include_optional=True)
            artifact = Artifact("compile_metrics.csv", content=content, fmt=ArtifactFormat.TEXT, flags=["metrics"])
            if name not in artifacts:
                artifacts[name] = []
            artifacts[name].append(artifact)
        return artifacts


class TargetPlatform(Platform):
    """Abstract target platform class."""

    def create_target(self, name):
        raise NotImplementedError

    @property
    def supports_flash(self):
        return True

    @property
    def supports_monitor(self):
        return True

    def flash(self, elf, target, timeout=120):
        raise NotImplementedError

    def monitor(self, target, timeout=60):
        raise NotImplementedError

    def run(self, elf, target, timeout=120):
        # Only allow one serial communication at a time
        with FileLock(Path(tempfile.gettempdir()) / "mlonmcu_serial.lock"):
            self.flash(elf, target, timeout=timeout)
            output = self.monitor(target, timeout=timeout)

        return output



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definitions for TFLMFramework."""

from pathlib import Path

from mlonmcu.flow.framework import Framework
from mlonmcu.flow.tflm import TFLMBackend


class TFLMFramework(Framework):
    """TFLM Framework specialization."""

    name = "tflm"

    FEATURES = {"muriscvnn", "cmsisnn"}

    DEFAULTS = {
        "optimized_kernel": None,
        "optimized_kernel_inc_dirs": [],
        "optimized_kernel_libs": [],
    }

    REQUIRED = {"tf.src_dir"}

    backends = TFLMBackend.registry

    def __init__(self, features=None, config=None):
        super().__init__(features=features, config=config)

    @property
    def tf_src(self):
        return Path(self.config["tf.src_dir"])

    @property
    def optimized_kernel(self):
        return self.config["optimized_kernel"]

    @property
    def optimized_kernel_libs(self):
        return self.config["optimized_kernel_libs"]

    @property
    def optimized_kernel_inc_dirs(self):
        return self.config["optimized_kernel_inc_dirs"]

    def get_platform_defs(self, platform):
        ret = super().get_platform_defs(platform)
        if self.optimized_kernel or self.optimized_kernel_inc_dirs or self.optimized_kernel_libs:
            if self.optimized_kernel:
                ret["TFLM_OPTIMIZED_KERNEL"] = self.optimized_kernel
            if self.optimized_kernel_inc_dirs:
                if isinstance(self.optimized_kernel_inc_dirs, list):
                    temp = r"\;".join(self.optimized_kernel_inc_dirs)
                else:
                    temp = self.optimized_kernel_inc_dirs
                ret["TFLM_OPTIMIZED_KERNEL_INCLUDE_DIR"] = temp
            if self.optimized_kernel_libs:
                if isinstance(self.optimized_kernel_libs, list):
                    temp = r"\;".join(self.optimized_kernel_libs)
                else:
                    temp = self.optimized_kernel_libs
                ret["TFLM_OPTIMIZED_KERNEL_LIB"] = temp
        ret["TF_DIR"] = str(self.tf_src)
        return ret



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU Vicuna Target definitions"""

import os
import re
from pathlib import Path
from tempfile import TemporaryDirectory
import time

from mlonmcu.logging import get_logger

# from mlonmcu.feature.features import SUPPORTED_TVM_BACKENDS
from mlonmcu.setup.utils import execute
from mlonmcu.target.common import cli
from mlonmcu.target.metrics import Metrics
from mlonmcu.target.bench import add_bench_metrics
from mlonmcu.setup import utils
from .riscv_vext_target import RVVTarget

logger = get_logger()

"""
NOTES
====

# select the core to use as main processor (defaults to Ibex)
CORE     ?= ibex
CORE_DIR := $(SIM_DIR)/../$(CORE)/

# memory initialization files
PROG_PATHS ?= progs.txt

# trace file
TRACE_FILE ?= sim_trace.csv
TRACE_SIGS ?= '*'

"""


class VicunaTarget(RVVTarget):
    """Target using a Pulpino-like VP running in the GVSOC simulator"""

    FEATURES = RVVTarget.FEATURES | {"log_instrs"}  # TODO: cache feature?

    DEFAULTS = {
        **RVVTarget.DEFAULTS,
        # vext config
        "vlen": 128,  # default value for hardware compilation, will be overwritten by -c vext.vlen
        "embedded_vext": True,
        "elen": 32,
        # riscv config
        "compressed": False,
        "atomic": False,
        "fpu": None,
        # processor config
        "core": "cv32e40x",  # TODO: also support ibex?
        # vector config
        "vproc_config": "compact",  # either compact, dual, triple, legacy, custom
        "vport_policy": None,  # only used for custom config
        "vmem_width": None,
        # "vreg_w": None,  -> determined by vext.vlen!
        "vproc_pipelines": None,
        # memory config
        "mem_width": 32,
        # "mem_size": 262144,
        "mem_size": 262144 * 2,
        "mem_latency": 1,
        # cache config
        "ic_size": 0,  # off
        "ic_line_width": 128,
        "dc_size": 0,  # off
        "dc_line_width": None,  # AUTO (2*VMEM_W)
        # trace config
        # TODO
        # testbench config
        "abort_cycles": 10000000,  # Used to detect freezes
        "extra_cycles": 4096,  # Number of remaining cycles after jump to reset vector
    }

    REQUIRED = RVVTarget.REQUIRED | {
        "vicuna.src_dir",  # for the bsp package
        "verilator.install_dir",  # for simulation
    }

    def __init__(self, name="vicuna", features=None, config=None):
        super().__init__(name, features=features, config=config)
        assert self.xlen == 32, "Vicuna target must have xlen=32"
        assert not self.enable_vext or self.embedded_vext, "Vicuna target only support embedded vector ext"
        self.prj_dir = None
        self.obj_dir = None

    @property
    def verilator_install_dir(self):
        return Path(self.config["verilator.install_dir"])

    @property
    def vicuna_src_dir(self):
        return Path(self.config["vicuna.src_dir"])

    @property
    def core(self):
        value = self.config["core"]
        return value

    @property
    def mem_width(self):
        value = self.config["mem_width"]
        return int(value) if value is not None else value

    @property
    def mem_size(self):
        value = self.config["mem_size"]
        return int(value) if value is not None else value

    @property
    def mem_latency(self):
        value = self.config["mem_latency"]
        return int(value) if value is not None else value

    @property
    def ic_size(self):
        value = self.config["ic_size"]
        return int(value) if value is not None else value

    @property
    def ic_line_width(self):
        value = self.config["ic_line_width"]
        return int(value) if value is not None else value

    @property
    def dc_size(self):
        value = self.config["dc_size"]
        return int(value) if value is not None else value

    @property
    def dc_line_width(self):
        value = self.config["dc_line_width"]
        return int(value) if value is not None else value

    @property
    def vproc_config(self):
        value = self.config["vproc_config"]
        return value

    @property
    def vport_policy(self):
        value = self.config["vport_policy"]
        return value

    @property
    def vmem_width(self):
        value = self.config["vmem_width"]
        return int(value) if value is not None else value

    @property
    def vproc_pipelines(self):
        value = self.config["vproc_pipelines"]
        return value

    @property
    def abort_cycles(self):
        value = self.config["abort_cycles"]
        return int(value)

    @property
    def extra_cycles(self):
        value = self.config["extra_cycles"]
        return int(value)

    def get_config_args(self):
        ret = []
        if self.core is not None:
            ret.append(f"CORE={self.core}")
        if self.mem_width is not None:
            ret.append(f"MEM_W={self.mem_width}")
        if self.mem_size is not None:
            ret.append(f"MEM_SZ={self.mem_size}")
        if self.mem_latency is not None:
            ret.append(f"MEM_LATENCY={self.mem_latency}")
        if self.ic_size is not None:
            ret.append(f"ICACHE_SZ={self.ic_size}")
        if self.ic_line_width is not None:
            ret.append(f"ICACHE_LINE_W={self.ic_line_width}")
        if self.dc_size is not None:
            ret.append(f"DCACHE_SZ={self.dc_size}")
        if self.dc_line_width is not None:
            ret.append(f"DCACHE_LINE_W={self.dc_line_width}")
        ret.append(f"VREG_W={self.vlen}")
        if self.vproc_config is not None:
            ret.append(f"VPROC_CONFIG={self.vproc_config}")
        if self.vport_policy is not None:
            ret.append(f"VPORT_POLICY={self.vport_policy}")
        if self.vmem_width is not None:
            ret.append(f"VMEM_W={self.vmem_width}")
        if self.vproc_pipelines is not None:
            ret.append(f"VPROC_PIPELINES={self.vproc_pipelines}")
        return ret

    def prepare_simulator(self, cwd=os.getcwd(), **kwargs):
        # populate the ara verilator testbench directory
        sim_dir = self.vicuna_src_dir / "sim"
        self.prj_dir = TemporaryDirectory()
        self.obj_dir = Path(self.prj_dir.name) / "obj_dir"
        env = os.environ.copy()
        orig_path = env["PATH"]
        # print("self.obj_dir", self.obj_dir)
        # input("000")
        env["PATH"] = f"{self.verilator_install_dir}/bin:{orig_path}"
        out = utils.make("verilator-version-check", env=env, cwd=sim_dir, **kwargs)
        out += utils.make(
            self.obj_dir / "Vvproc_top.mk",
            f"PROJ_DIR={self.prj_dir.name}",
            f"SIM_ABORT_CYCLES={self.abort_cycles}",
            *self.get_config_args(),
            env=env,
            cwd=sim_dir,
            **kwargs,
        )
        # print("self.obj_dir", self.obj_dir)
        # input("111")
        out += utils.make("-f", self.obj_dir / "Vvproc_top.mk", "Vvproc_top", env=env, cwd=self.obj_dir, **kwargs)
        return out

    def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
        """Use target to execute an executable with given arguments"""
        vicuna_exe = self.obj_dir / "Vvproc_top"
        if len(self.extra_args) > 0:
            assert False, "Vicuna TB does not allow cmdline arguments"

        assert self.prj_dir is not None, "Not prepared?"
        path_file = f"{program}.path"

        trace_file = "trace.csv"
        # trace_vcd = "vcd"
        # trace_fst = "fst"
        vicuna_args = [
            path_file,
            str(self.mem_width),
            str(self.mem_size),
            str(self.mem_latency),
            str(self.extra_cycles),
            trace_file,
            # trace_vcd,
        ]
        env = os.environ.copy()
        out = execute(
            vicuna_exe,
            *vicuna_args,
            env=env,
            cwd=cwd,
            *args,
            **kwargs,
        )
        # print("prj", self.prj_dir)
        # print("out", out)
        # input("!")
        self.prj_dir.cleanup()
        self.obj_dir = None
        return out, []

    def parse_stdout(self, out, metrics, exit_code=0):
        add_bench_metrics(out, metrics, exit_code != 0)
        sim_insns = re.search(r"(\d*) cycles", out)
        if sim_insns:
            sim_insns = int(float(sim_insns.group(1)))
            metrics.add("Simulated Instructions", sim_insns, True)

    def parse_exit(self, out):
        exit_code = super().parse_exit(out)
        if exit_code is None:
            if "EXCEPTION: Illegal instruction at" in out:
                exit_code = -4
            elif "WARNING: memory interface inactive for" in out:
                exit_code = -3
            elif "Program finish." not in out:
                exit_code = -2  # did not finish
            # elif "EXCEP" in out:
            #     exit_code = -1  # unhandled exception
        return exit_code

    def get_metrics(self, elf, directory, *args, handle_exit=None):
        out = ""
        artifacts = []

        def _handle_exit(code, out=None):
            assert out is not None
            temp = self.parse_exit(out)
            # TODO: before or after?
            if temp is None:
                temp = code
            if handle_exit is not None:
                temp = handle_exit(temp, out=out)
            return temp

        if self.print_outputs:
            self.prepare_simulator(cwd=directory, live=True)
        else:
            self.prepare_simulator(cwd=directory, live=False, print_func=lambda *args, **kwargs: None)
        simulation_start = time.time()
        if self.print_outputs:
            out_, artifacts_ = self.exec(elf, *args, cwd=directory, live=True, handle_exit=_handle_exit)
        else:
            out_, artifacts_ = self.exec(
                elf, *args, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=_handle_exit
            )
        out += out_
        artifacts += artifacts_
        simulation_end = time.time()
        exit_code = 0
        metrics = Metrics()
        self.parse_stdout(out, metrics, exit_code=exit_code)
        wall = simulation_end - simulation_start
        # metrics.add("Wallclock time", wall, True)
        if metrics.has("Total Instructions"):
            instructions = metrics.get("Total Instructions")
            if instructions > 0:
                # Warning: Total Instructions != Simulated Instructions
                mips = (instructions / wall) / 1e6
                metrics.add("MIPS", mips, True)
        return metrics, out, artifacts

    def get_target_system(self):
        return self.name

    def get_platform_defs(self, platform):
        ret = super().get_platform_defs(platform)
        ret["VICUNA_DIR"] = self.vicuna_src_dir
        return ret

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = super().get_backend_config(
            backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
        )
        return ret


if __name__ == "__main__":
    cli(target=VicunaTarget)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU Corstone300 Target definitions"""

import os
import re
from pathlib import Path

from mlonmcu.logging import get_logger
from mlonmcu.feature.features import SUPPORTED_TVM_BACKENDS
from mlonmcu.setup.utils import execute
from mlonmcu.config import str2bool
from mlonmcu.target import Target
from mlonmcu.target.common import cli
from mlonmcu.target.metrics import Metrics
from .util import resolve_cpu_features

logger = get_logger()


class Corstone300Target(Target):
    """Target using an ARM FVP (fixed virtual platform) based on a Cortex M55 with EthosU support"""

    FEATURES = {"ethosu", "arm_mvei", "arm_dsp"}

    DEFAULTS = {
        **Target.DEFAULTS,
        "model": "cortex-m55",  # Options: cortex-m{0,0plus,1,3,4,7,23,33,35p,55} (Frequency is fixed at 25MHz)
        # Warning: FVP is still M55 based!
        "timeout_sec": 0,  # disabled
        "enable_ethosu": False,
        "enable_fpu": True,
        "enable_mvei": False,  # unused
        "enable_dsp": False,  # unused
        "ethosu_num_macs": 256,
        "extra_args": "",
    }
    REQUIRED = {
        "corstone300.exe",
        "cmsis.dir",
        "cmsisnn.dir",
        "ethosu_platform.dir",
        "arm_gcc.install_dir",
    }  # Actually cmsisnn.dir points to the root CMSIS_5 directory

    def __init__(self, name="corstone300", features=None, config=None):
        super().__init__(name, features=features, config=config)

    @property
    def model(self):
        return self.config["model"]

    @property
    def enable_ethosu(self):
        value = self.config["enable_ethosu"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_fpu(self):
        value = self.config["enable_fpu"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_mvei(self):
        value = self.config["enable_mvei"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def enable_dsp(self):
        value = self.config["enable_dsp"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def ethosu_num_macs(self):
        return int(self.config["ethosu_num_macs"])

    @property
    def fvp_exe(self):
        return Path(self.config["corstone300.exe"])

    @property
    def gcc_prefix(self):
        return str(self.config["arm_gcc.install_dir"])

    @property
    def cmsis_dir(self):
        return Path(self.config["cmsis.dir"])

    @property
    def cmsisnn_dir(self):
        return Path(self.config["cmsisnn.dir"])

    @property
    def ethosu_platform_dir(self):
        return Path(self.config["ethosu_platform.dir"])

    @property
    def extra_args(self):
        return str(self.config["extra_args"])

    @property
    def timeout_sec(self):
        # 0 = off
        return int(self.config["timeout_sec"])

    def get_default_fvp_args(self):
        return [
            "-C",
            "mps3_board.visualisation.disable-visualisation=1",
            "-C",
            "mps3_board.telnetterminal0.start_telnet=0",
            "-C",
            'mps3_board.uart0.out_file="-"',
            "-C",
            "mps3_board.uart0.unbuffered_output=1",
            "-C",
            'mps3_board.uart0.shutdown_tag="EXITTHESIM"',
            "-C",
            "cpu0.CFGDTCMSZ=15",  # ?
            "-C",
            "cpu0.CFGITCMSZ=15",  # ?
            "-C",
            f"cpu0.FPU={int(self.enable_fpu)}",
            "-C",
            "cpu0.MVE={}".format(
                2 if self.enable_mvei and self.enable_fpu else (1 if self.enable_mvei and not self.enable_fpu else 0)
            ),
        ]

    def get_ethosu_fvp_args(self):
        return [
            "-C",
            f"ethosu.num_macs={self.ethosu_num_macs}",
            "-C",
            'ethosu.extra_args="--fast"',
        ]

    def exec(self, program, *args, cwd=os.getcwd(), **kwargs):
        """Use target to execute a executable with given arguments"""
        fvp_args = []
        fvp_args.extend(self.get_default_fvp_args())
        if self.enable_ethosu:
            fvp_args.extend(self.get_ethosu_fvp_args())
        if self.timeout_sec > 0:
            fvp_args.extend(["--timelimit", str(self.timeout_sec)])
        if len(self.extra_args) > 0:
            fvp_args.extend(self.extra_args.split(" "))

        if "ethosu" in [feature.name for feature in self.features]:  # TODO: remove this
            raise NotImplementedError

        ret = execute(
            self.fvp_exe.resolve(),
            *fvp_args,
            program,
            *args,
            cwd=cwd,
            **kwargs,
        )
        return ret

    def parse_stdout(self, out, handle_exit=None):
        exit_match = re.search(r"Application exit code: (.*)\.", out)
        if exit_match:
            exit_code = int(exit_match.group(1))
            if handle_exit is not None:
                exit_code = handle_exit(exit_code)
            if exit_code != 0:
                logger.error("Execution failed - " + out)
                raise RuntimeError(f"unexpected exit code: {exit_code}")
        cpu_cycles = re.search(r"Total Cycles: (.*)", out)

        if not cpu_cycles:
            if exit == 0:
                logger.warning("unexpected script output (cycles)")
            cycles = None
        else:
            cycles = int(float(cpu_cycles.group(1)))
        # mips = None  # TODO: parse mips?
        return cycles

    def get_metrics(self, elf, directory, *args, handle_exit=None):
        out = ""
        if self.print_outputs:
            out += self.exec(elf, *args, cwd=directory, live=True, handle_exit=handle_exit)
        else:
            out += self.exec(
                elf, *args, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=handle_exit
            )
        cycles = self.parse_stdout(out, handle_exit=handle_exit)

        metrics = Metrics()
        metrics.add("Cycles", cycles)

        return metrics, out, []

    def get_platform_defs(self, platform):
        ret = super().get_platform_defs(platform)
        ret["CMSIS_DIR"] = self.cmsis_dir
        ret["CMSISNN_DIR"] = self.cmsisnn_dir
        ret["ETHOSU_PLATFORM_DIR"] = self.ethosu_platform_dir
        ret["ARM_COMPILER_PREFIX"] = self.gcc_prefix
        cpu, float_abi, fpu = resolve_cpu_features(
            self.model, enable_fp=self.enable_fpu, enable_dsp=self.enable_dsp, enable_mve=self.enable_mvei
        )
        ret["ARM_CPU"] = cpu
        ret["ARM_FLOAT_ABI"] = float_abi
        ret["ARM_FPU"] = fpu
        return ret

    def get_arch(self):
        return "arm"  # TODO: use proper mapping (v6, v7, v8, v8.1...)

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = {}
        cpu, _, _ = resolve_cpu_features(
            self.model, enable_fp=self.enable_fpu, enable_dsp=self.enable_dsp, enable_mve=self.enable_mvei
        )
        if backend in SUPPORTED_TVM_BACKENDS:
            ret = {
                # "target_march": self.get_arch(),
                "target_mtriple": "arm-none-eabi",
                "target_mcpu": self.model,
                # "target_mattr": "?",
                # "target_mabi": self.abi,
                "target_model": f"{self.name}-{cpu}",
            }
            if optimized_schedules:
                ret.update(
                    {
                        "target_device": "arm_cpu",
                    }
                )

                if optimized_layouts:
                    if self.enable_dsp:
                        ret.update({"desired_layout": "NHWC:HWOI"})
        return ret


if __name__ == "__main__":
    cli(target=Corstone300Target)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path
import pytest
from mlonmcu.context.context import MlonMcuContext


def create_minimal_environment_yaml(path):
    dirname = path.parent.absolute()
    with open(path, "w") as f:
        f.write(f"---\nhome: {dirname}")  # Use defaults


def create_invalid_environment_yaml(path):
    dirname = path.parent.absolute()
    with open(path, "w") as f:
        f.write(f"---\nhome: {dirname}")  # Use defaults


# def test_resolve_environment_file():
# def test_resolve_environment_file_by_name():
# def test_resolve_environment_file_by_file():
# def test_resolve_environment_file_by_dir():
# def test_resolve_environment_file_by_cwd():
# def test_resolve_environment_file_by_env():
# def test_resolve_environment_file_by_default():
# def test_load_recent_sessions():
# def test_create_session():
# def test_load_cache():
# def test_get_session():


def test_open_context(monkeypatch, fake_environment_directory: Path, fake_config_home: Path):
    monkeypatch.chdir(fake_environment_directory)
    create_minimal_environment_yaml(fake_environment_directory / "environment.yml")
    ctx = None
    with MlonMcuContext() as context:
        assert context
        ctx = context
    assert ctx.is_clean


# def test_open_context_by_env(monkeypatch, fake_environment_directory: Path, fake_config_home: Path):
#     monkeypatch.chdir(fake_environment_directory)
#     create_minimal_environment_yaml(fake_environment_directory / "environment.yml")
#     with mlonmcu.context.MlonMcuContext() as context:
#         assert context
#
# def test_open_context_by_default(monkeypatch, fake_environment_directory: Path, fake_config_home: Path):
#     monkeypatch.chdir(fake_environment_directory)
#     create_minimal_environment_yaml(fake_environment_directory / "environment.yml")
#     with mlonmcu.context.MlonMcuContext() as context:
#         assert context
#
# def test_open_context_by_path(monkeypatch, fake_environment_directory: Path, fake_config_home: Path):
#     monkeypatch.chdir(fake_environment_directory)
#     create_minimal_environment_yaml(fake_environment_directory / "environment.yml")
#     with mlonmcu.context.MlonMcuContext() as context:
#         assert context
#
# def test_open_context_by_name(monkeypatch, fake_environment_directory: Path, fake_config_home: Path):
#     monkeypatch.chdir(fake_environment_directory)
#     create_minimal_environment_yaml(fake_environment_directory / "environment.yml")
#     with mlonmcu.context.MlonMcuContext() as context:
#         assert context


def test_reuse_context(monkeypatch, fake_environment_directory: Path, fake_config_home: Path):
    monkeypatch.chdir(fake_environment_directory)
    create_minimal_environment_yaml(fake_environment_directory / "environment.yml")
    with MlonMcuContext() as context:
        assert context
    with MlonMcuContext() as context2:
        assert context2


def test_reuse_context_locked(monkeypatch, fake_environment_directory: Path, fake_config_home: Path):
    monkeypatch.chdir(fake_environment_directory)
    create_minimal_environment_yaml(fake_environment_directory / "environment.yml")
    with MlonMcuContext() as context:
        assert context
    with MlonMcuContext() as context2:
        assert context2


def test_nest_context_read_after_read(monkeypatch, fake_environment_directory: Path, fake_config_home: Path):
    monkeypatch.chdir(fake_environment_directory)
    create_minimal_environment_yaml(fake_environment_directory / "environment.yml")
    with MlonMcuContext(deps_lock="read") as context:
        assert context
        with MlonMcuContext(deps_lock="read") as context2:
            assert context2


def test_nest_context_read_after_write(monkeypatch, fake_environment_directory: Path, fake_config_home: Path):
    monkeypatch.chdir(fake_environment_directory)
    create_minimal_environment_yaml(fake_environment_directory / "environment.yml")

    with MlonMcuContext(deps_lock="write") as context:
        assert context
        with pytest.raises(RuntimeError, match=r".*could\ not\ be\ acquired\..*"):
            with MlonMcuContext(deps_lock="read") as context2:
                assert context2


def test_nest_context_write_after_read(monkeypatch, fake_environment_directory: Path, fake_config_home: Path):
    monkeypatch.chdir(fake_environment_directory)
    create_minimal_environment_yaml(fake_environment_directory / "environment.yml")

    with MlonMcuContext(deps_lock="write") as context:
        assert context
        with pytest.raises(RuntimeError, match=r".*could\ not\ be\ acquired\..*"):
            with MlonMcuContext(deps_lock="read") as context2:
                assert context2


def test_nest_context_write_after_write(monkeypatch, fake_environment_directory: Path, fake_config_home: Path):
    monkeypatch.chdir(fake_environment_directory)
    create_minimal_environment_yaml(fake_environment_directory / "environment.yml")

    with MlonMcuContext(deps_lock="write") as context:
        assert context
        with pytest.raises(RuntimeError, match=r".*could\ not\ be\ acquired\..*"):
            with MlonMcuContext(deps_lock="read") as context2:
                assert context2



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from .mlif import MlifPlatform
from .espidf import EspIdfPlatform
from .zephyr import ZephyrPlatform
from .tvm import TvmPlatform
from .microtvm import MicroTvmPlatform

# from .arduino import ArduinoPlatform


PLATFORM_REGISTRY = {}


def register_platform(platform_name, p, override=False):
    global PLATFORM_REGISTRY

    if platform_name in PLATFORM_REGISTRY and not override:
        raise RuntimeError(f"Platform {platform_name} is already registered")
    PLATFORM_REGISTRY[platform_name] = p


def get_platforms():
    return PLATFORM_REGISTRY


register_platform("mlif", MlifPlatform)
register_platform("espidf", EspIdfPlatform)
register_platform("zephyr", ZephyrPlatform)
register_platform("tvm", TvmPlatform)
register_platform("microtvm", MicroTvmPlatform)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MicroTVM Target Platform"""
from mlonmcu.config import str2bool
from ..tvm.tvm_target_platform import TvmTargetPlatform
from .microtvm_base_platform import (
    filter_project_options,
    get_project_option_args,
)  # TODO: move to utils
from mlonmcu.flow.tvm.backend.tvmc_utils import (
    get_bench_tvmc_args,
    get_data_tvmc_args,
    # get_rpc_tvmc_args,
)
from mlonmcu.logging import get_logger

logger = get_logger()


class MicroTvmTargetPlatform(TvmTargetPlatform):
    """MicroTVM target platform class."""

    FEATURES = (
        # TvmTargetPlatform.FEATURES +
        set()
        # Warning: benchamrk and profile not supported!
    )

    DEFAULTS = {
        **TvmTargetPlatform.DEFAULTS,
        "experimental_tvmc_print_time": False,
        "skip_flash": False,
        # Warning: contains configs not supported by microtvm
    }

    @property
    def experimental_tvmc_print_time(self):
        value = self.config["experimental_tvmc_print_time"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def skip_flash(self):
        value = self.config["skip_flash"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def invoke_tvmc_micro_flash(self, target=None, list_options=False, **kwargs):
        all_args = []
        all_args.append(self.project_dir)
        template_args = self.get_template_args(target)
        all_args.extend(template_args)
        return self.invoke_tvmc_micro("flash", *all_args, target=target, list_options=list_options, **kwargs)

    def invoke_tvmc_micro_run(self, *args, target=None, list_options=False, **kwargs):
        all_args = []
        all_args.extend(args)
        all_args.append(self.project_dir)
        all_args.extend(["--device", "micro"])
        if list_options:
            all_args.append("--list-options")
        else:
            options = filter_project_options(
                self.collect_available_project_options("run", target=target), target.get_project_options()
            )
            all_args.extend(get_project_option_args("run", options))
        return self.invoke_tvmc("run", *all_args, target=target, **kwargs)

    def invoke_tvmc_run(self, *args, target=None):
        return self.invoke_tvmc_micro_run(*args, target=target)

    def get_tvmc_run_args(self):
        if self.use_rpc:
            raise RuntimeError("RPC is only supported for tuning with microtvm platform")
        if self.profile:
            assert (
                self.experimental_tvmc_print_time
            ), "MicroTVM profiling is only supported in environments with microtvm.experimental_tvmc_print_time=1"
        ret = [
            *get_data_tvmc_args(
                mode=self.fill_mode, ins_file=self.ins_file, outs_file=self.outs_file, print_top=self.print_top
            ),
            *get_bench_tvmc_args(
                print_time=self.experimental_tvmc_print_time and not self.profile,
                profile=self.profile and self.experimental_tvmc_print_time,
                end_to_end=False,
                repeat=self.repeat if self.experimental_tvmc_print_time else None,
                number=self.number if self.experimental_tvmc_print_time else None,
            ),
            # *get_rpc_tvmc_args(self.use_rpc, self.rpc_key, self.rpc_hostname, self.rpc_port),
        ]
        return ret

    def flash(self, elf, target, timeout=120):
        # Ignore elf, as we use self.project_dir instead
        # TODO: add alternative approach which allows passing elf instead
        if elf is not None:
            logger.debug("Ignoring ELF file for microtvm platform")
        # TODO: implement timeout
        logger.debug("Flashing target software using MicroTVM ProjectAPI")
        output = self.invoke_tvmc_micro_flash(target=target)
        return output

    def monitor(self, target, timeout=60):
        raise NotImplementedError

    def run(self, elf, target, timeout=120):
        # TODO: implement timeout
        output = ""
        if not self.skip_flash:
            output += self.flash(elf, target)
        run_args = self.get_tvmc_run_args()
        output += self.invoke_tvmc_run(*run_args, target=target)

        return output



from multiprocessing import Process, Queue


def exec_timeout(timeout, func, *args, **kwargs):
    retQueue = Queue()

    def wrap_call():
        try:
            retQueue.put(func(*args, **kwargs))
        except Exception as e:
            retQueue.put(e)

    p = Process(target=wrap_call)
    p.start()
    p.join(timeout)
    if p.is_alive():
        p.terminate()
        p.join(1)
        if p.is_alive():
            p.kill()
            p.join()
        raise TimeoutError("Function did not complete within the timeout time")

    ret = retQueue.get()
    if isinstance(ret, Exception):
        raise ret
    return ret



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""ESP-IDF Platform"""

import os
import time
import signal
import shutil
import tempfile
import subprocess
from pathlib import Path
import pkg_resources


from mlonmcu.setup import utils
from mlonmcu.artifact import Artifact, ArtifactFormat
from mlonmcu.logging import get_logger
from mlonmcu.target.target import Target
from mlonmcu.config import str2bool

from ..platform import CompilePlatform, TargetPlatform
from .espidf_target import create_espidf_platform_target, get_espidf_platform_targets

logger = get_logger()


def get_project_template(name="project"):
    espidf_templates = pkg_resources.resource_listdir("mlonmcu", os.path.join("..", "resources", "platforms", "espidf"))
    if name not in espidf_templates:
        return None
    fname = pkg_resources.resource_filename("mlonmcu", os.path.join("..", "resources", "platforms", "espidf", name))
    return fname


class EspIdfPlatform(CompilePlatform, TargetPlatform):
    """ESP-IDF Platform class."""

    FEATURES = CompilePlatform.FEATURES | TargetPlatform.FEATURES | {"benchmark"}

    DEFAULTS = {
        **CompilePlatform.DEFAULTS,
        **TargetPlatform.DEFAULTS,
        "project_template": None,
        "project_dir": None,
        "port": None,
        "baud": 115200,
        "use_idf_monitor": True,
        "wait_for_user": True,
        "flash_only": False,
    }

    REQUIRED = {"espidf.install_dir", "espidf.src_dir"}

    def __init__(self, features=None, config=None):
        super().__init__(
            "espidf",
            features=features,
            config=config,
        )
        self.tempdir = None
        self.project_name = "app"
        self.project_dir = None

    @property
    def espidf_install_dir(self):
        return Path(self.config["espidf.install_dir"])

    @property
    def espidf_src_dir(self):
        return Path(self.config["espidf.src_dir"])

    @property
    def idf_exe(self):
        return self.espidf_src_dir / "tools" / "idf.py"

    @property
    def use_idf_monitor(self):
        value = self.config["use_idf_monitor"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def wait_for_user(self):
        value = self.config["wait_for_user"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def flash_only(self):
        # TODO: get rid of this
        value = self.config["flash_only"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def invoke_idf_exe(self, *args, **kwargs):
        env = os.environ.copy()
        env["IDF_PATH"] = str(self.espidf_src_dir)
        env["IDF_TOOLS_PATH"] = str(self.espidf_install_dir)
        cmd = (
            ". "
            + str(self.espidf_src_dir / "export.sh")
            + f" && {self.idf_exe} "
            # + f" > /dev/null && {self.idf_exe} "
            + " ".join([str(arg) for arg in args])
        )
        out = utils.exec_getout(
            cmd, shell=True, env=env, **kwargs, executable="/bin/bash"
        )  # TODO: using shell=True is insecure but right now we can not avoid it?
        return out

    def init_directory(self, path=None, context=None):
        if self.project_dir is not None:
            self.project_dir.mkdir(exist_ok=True)
            logger.debug("Project directory already initialized")
            return
        dir_name = self.name
        if path is not None:
            self.project_dir = Path(path)
        elif self.config["project_dir"] is not None:
            self.project_dir = Path(self.config["project_dir"])
        else:
            if context:
                assert "temp" in context.environment.paths
                self.project_dir = (
                    context.environment.paths["temp"].path / dir_name
                )  # TODO: Need to lock this for parallel builds
            else:
                logger.debug(
                    "Creating temporary directory because no context was available "
                    "and 'espidf.project_dir' was not supplied"
                )
                self.tempdir = tempfile.TemporaryDirectory()
                self.project_dir = Path(self.tempdir.name) / dir_name
                logger.debug("Temporary project directory: %s", self.project_dir)
        self.project_dir.mkdir(exist_ok=True)

    def get_supported_targets(self):
        text = self.invoke_idf_exe("--list-targets", live=self.print_outputs)
        # Warning: This will fail if a python executable is NOT available in the system. Aliasing
        # python3 to python will not work. Not sure how this would handle a system which only has python2 installed?
        target_names = text.split("\n")

        def filter_names(x):
            ret = []
            for x in reversed(x):
                if len(x) > 0:
                    ret.append(x)
                else:
                    if len(ret) > 0:
                        break
                    else:
                        continue
            return reversed(ret)

        target_names = filter_names(target_names)

        return [name for name in target_names if len(name) > 0 and " " not in name]

    def create_target(self, name):
        assert name in self.get_supported_targets(), f"{name} is not a valid ESP-IDF target"
        targets = get_espidf_platform_targets()
        if name in targets:
            base = targets[name]
        else:
            base = Target
        return create_espidf_platform_target(name, self, base=base)

    @property
    def project_template(self):
        return self.config["project_template"]

    @property
    def port(self):
        return self.config["port"]

    @property
    def baud(self):
        return self.config["baud"]

    def close(self):
        if self.tempdir:
            self.tempdir.cleanup()

    def check(self):
        if not shutil.which(self.idf_exe):
            raise RuntimeError(f"It seems like '{self.idf_exe}' is not available. Make sure to setup your environment!")

    def prepare(self, target, src):
        self.init_directory()
        self.check()
        template_dir = self.project_template
        if template_dir is None:
            template_dir = get_project_template()
        else:
            template_dir = Path(template_dir)
            if not template_dir.is_dir():
                template_dir = get_project_template(name=str(template_dir))
        assert template_dir is not None, f"Provided project template does not exists: {template_dir}"
        shutil.copytree(template_dir, self.project_dir, dirs_exist_ok=True)

        def write_defaults(filename):
            defs = self.definitions
            with open(filename, "w", encoding="utf-8") as f:
                f.write("CONFIG_PARTITION_TABLE_SINGLE_APP_LARGE=y\n")
                if self.debug:
                    f.write("CONFIG_OPTIMIZATION_LEVEL_DEBUG=y\n")
                    f.write("CONFIG_COMPILER_OPTIMIZATION_LEVEL_DEBUG=y\n")
                else:
                    # Trying to reduce the binary size as much as possible
                    # (https://docs.espressif.com/projects/esp-idf/en/latest/esp32/api-guides/performance/size.html)
                    f.write("CONFIG_BOOTLOADER_LOG_LEVEL_NONE=y\n")
                    f.write("CONFIG_BOOTLOADER_LOG_LEVEL=0\n")
                    f.write("CONFIG_BOOT_ROM_LOG_ALWAYS_OFF=y\n")
                    f.write("CONFIG_COMPILER_OPTIMIZATION_ASSERTIONS_DISABLE=y\n")
                    f.write("CONFIG_COMPILER_OPTIMIZATION_ASSERTION_LEVEL=0\n")
                    f.write("CONFIG_COMPILER_OPTIMIZATION_CHECKS_SILENT=y\n")
                    f.write("CONFIG_HAL_DEFAULT_ASSERTION_LEVEL=0\n")
                    f.write("CONFIG_LOG_DEFAULT_LEVEL_NONE=y\n")
                    f.write("CONFIG_LOG_DEFAULT_LEVEL=0\n")
                    f.write("CONFIG_LOG_MAXIMUM_LEVEL=0\n")
                    f.write("CONFIG_NEWLIB_NANO_FORMAT=y\n")
                    f.write("CONFIG_COMPILER_OPTIMIZATION_LEVEL_RELEASE=y\n")
                    f.write("CONFIG_OPTIMIZATION_LEVEL_RELEASE=y\n")
                    optimize_for_size = True
                    if optimize_for_size:
                        f.write("CONFIG_COMPILER_OPTIMIZATION_SIZE=y\n")
                    else:
                        f.write("CONFIG_COMPILER_OPTIMIZATION_PERF=y\n")
                watchdog_sec = 60
                f.write(f"CONFIG_ESP_TASK_WDT_TIMEOUT_S={watchdog_sec}\n")
                f.write(f'CONFIG_MLONMCU_CODEGEN_DIR="{src}"\n')
                for key, value in defs.items():
                    if isinstance(value, bool):
                        value = "y" if value else "n"
                    else:
                        value = f'"{value}"'
                    f.write(f"CONFIG_{key}={value}\n")

        write_defaults(self.project_dir / "sdkconfig.defaults")
        idfArgs = [
            "-C",
            self.project_dir,
            "set-target",
            target.name,
        ]
        out = self.invoke_idf_exe(*idfArgs, live=self.print_outputs)
        return out

    def get_idf_cmake_args(self):
        cmake_defs = {"CMAKE_BUILD_TYPE": "Debug" if self.debug else "Release"}
        return [f"-D{key}={value}" for key, value in cmake_defs.items()]

    def compile(self, target, src=None):
        out = ""
        # TODO: build with cmake options
        out += self.prepare(target, src)
        # TODO: support self.num_threads (e.g. patch esp-idf)
        idfArgs = [
            "-C",
            self.project_dir,
            *self.get_idf_cmake_args(),
            "build",
        ]
        out += self.invoke_idf_exe(*idfArgs, live=self.print_outputs)
        return out

    def generate(self, src, target, model=None):
        artifacts = []
        out = self.compile(target, src=src)
        elf_name = self.project_name + ".elf"
        elf_file = self.project_dir / "build" / elf_name
        # TODO: just use path instead of raw data?
        if self.tempdir:
            # Warning: The porject context will get destroyed afterwards wehen using  a temporory directory
            with open(elf_file, "rb") as handle:
                data = handle.read()
                artifact = Artifact("generic_mlif", raw=data, fmt=ArtifactFormat.RAW)
                artifacts.append(artifact)
        else:
            artifact = Artifact(elf_name, path=elf_file, fmt=ArtifactFormat.PATH)
            artifacts.append(artifact)
        metrics = self.get_metrics(elf_file)
        stdout_artifact = Artifact(
            "espidf_out.log", content=out, fmt=ArtifactFormat.TEXT  # TODO: split into one file per command
        )  # TODO: rename to tvmaot_out.log?
        artifacts.append(stdout_artifact)
        return {"default": artifacts}, {"default": metrics}

    def get_idf_serial_args(self, monitor=False):
        args = []
        if self.port:
            args.extend(["-p", self.port])
        if self.baud:
            args.extend(["-B" if monitor else "-b", self.baud])
        return args

    def flash(self, elf, target, timeout=120):
        # Ignore elf, as we use self.project_dir instead
        # TODO: add alternative approach which allows passing elf instead
        if elf is not None:
            logger.debug("Ignoring ELF file for espidf platform")
        # TODO: implement timeout
        # TODO: make sure that already compiled? -> error or just call compile routine?
        if self.wait_for_user:  # INTERACTIVE
            answer = input(
                f"Make sure that the device '{target.name}' is connected before you press [Enter]"
                + " (Type 'Abort' to cancel)"
            )
            if answer.lower() == "abort":
                return ""
        logger.debug("Flashing target software")

        idfArgs = [
            "-C",
            self.project_dir,
            *self.get_idf_cmake_args(),
            "flash",
            *self.get_idf_serial_args(),
        ]
        self.invoke_idf_exe(*idfArgs, live=self.print_outputs)

    def monitor(self, target, timeout=60):
        if self.flash_only:
            return ""

        if self.use_idf_monitor:

            def _kill_monitor():
                import psutil

                for proc in psutil.process_iter():
                    # check whether the process name matches
                    cmdline = " ".join(proc.cmdline())
                    if "idf_monitor.py" in cmdline:  # TODO: do something less "dangerous"?
                        proc.kill()

            def _monitor_helper(*args, verbose=False, start_match=None, end_match=None, timeout=60):
                # start_match and end_match are inclusive
                if timeout:
                    pass  # TODO: implement timeout
                found_start = start_match is None
                logger.debug("- Executing: %s", str(args))
                outStr = ""
                env = os.environ.copy()
                env["IDF_PATH"] = str(self.espidf_src_dir)
                env["IDF_TOOLS_PATH"] = str(self.espidf_install_dir)
                cmd = (
                    ". "
                    + str(self.espidf_src_dir / "export.sh")
                    + f"> /dev/null && {self.idf_exe} "
                    + " ".join([str(arg) for arg in args])
                )
                process = subprocess.Popen(
                    cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, executable="/bin/bash", env=env
                )
                try:
                    exit_code = None
                    for line in process.stdout:
                        new_line = line.decode(errors="replace")
                        if verbose:
                            print(new_line.replace("\n", ""))
                        if start_match and start_match in new_line:
                            outStr = new_line
                            found_start = True
                        else:
                            outStr = outStr + new_line
                        if found_start:
                            if end_match and end_match in new_line:
                                _kill_monitor()
                                process.terminate()
                                exit_code = 0
                    while exit_code is None:
                        exit_code = process.poll()
                    if not verbose and exit_code != 0:
                        logger.error(outStr)
                    assert exit_code == 0, "The process returned an non-zero exit code {}! (CMD: `{}`)".format(
                        exit_code, cmd
                    )
                except KeyboardInterrupt:
                    logger.debug("Interrupted subprocess. Sending SIGINT signal...")
                    _kill_monitor()
                    pid = process.pid
                    os.kill(pid, signal.SIGINT)
                os.system("reset")
                return outStr

            logger.debug("Monitoring target software")
            # TODO: implement timeout
            idfArgs = [
                "-C",
                self.project_dir,
                *self.get_idf_cmake_args(),
                "monitor",
                *self.get_idf_serial_args(monitor=True),
            ]
            return _monitor_helper(
                *idfArgs,
                verbose=self.print_outputs,
                start_match="MLonMCU: START",
                end_match="MLonMCU: STOP",
                timeout=timeout,
            )
        else:
            port = target.port
            if port is None:
                # Falling back to 'espidf.baud'
                assert self.port is not None, f"If using custom serial monitor, please provide '{target.name}.port'"
                port = self.port
            baud = target.baud
            if baud is None:
                # Falling back to 'espidf.baud'
                assert self.baud is not None, f"If using custom serial monitor, please provide '{target.name}.baud'"
                baud = self.baud

            def _monitor_helper2(port, baud, verbose=False, start_match=None, end_match=None, timeout=60):
                # Local import to make this only required for real HW targets
                import serial

                # start_match and end_match are inclusive
                found_start = start_match is None
                outStr = ""
                if timeout:
                    pass  # TODO: implement timeout
                # The following is a custom initialization sequence inspired by
                # (https://github.com/espressif/esp-idf/blob/master/tools/idf_monitor_base/serial_reader.py)
                high = False
                low = True
                ser = serial.Serial(port, baud)
                ser.close()
                ser.dtr = False
                ser.rts = False
                time.sleep(1)
                ser.dtr = low
                ser.rts = high
                ser.dtr = ser.dtr
                ser.open()
                ser.dtr = high
                ser.rts = low
                ser.dtr = ser.dtr
                time.sleep(0.002)
                ser.rts = high
                ser.dtr = ser.dtr
                try:
                    while True:
                        try:
                            ser_bytes = ser.readline()
                            new_line = ser_bytes.decode("utf-8", errors="replace")
                            if verbose:
                                print(new_line.replace("\n", ""))
                            if start_match and start_match in new_line:
                                outStr = new_line
                                found_start = True
                            else:
                                outStr = outStr + new_line
                            if found_start:
                                if end_match and end_match in new_line:
                                    break
                        except KeyboardInterrupt:
                            logger.warning("Stopped processing serial port (KeyboardInterrupt)")
                finally:
                    ser.close()
                return outStr

            logger.debug("Monitoring target software")
            # TODO: implement timeout
            return _monitor_helper2(
                port,
                baud,
                verbose=self.print_outputs,
                start_match="MLonMCU: START",
                end_match="MLonMCU: STOP",
                timeout=timeout,
            )



from .etiss_pulpino import EtissPulpinoTarget
from .etiss import EtissTarget
from .spike import SpikeTarget
from .ovpsim import OVPSimTarget
from .corev_ovpsim import COREVOVPSimTarget
from .riscv_qemu import RiscvQemuTarget
from .gvsoc_pulp import GvsocPulpTarget
from .ara import AraTarget
from .ara_rtl import AraRtlTarget
from .cv32e40p import CV32E40PTarget
from .vicuna import VicunaTarget

__all__ = [
    "EtissPulpinoTarget",
    "EtissTarget",
    "SpikeTarget",
    "OVPSimTarget",
    "COREVOVPSimTarget",
    "RiscvQemuTarget",
    "GvsocPulpTarget",
    "AraTarget",
    "AraRtlTarget",
    "CV32E40PTarget",
    "VicunaTarget",
]



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest
import mock
import mlonmcu.platform.platform
from mlonmcu.platform.platform import TargetPlatform
import mlonmcu.platform.lookup
from mlonmcu.platform import register_platform, get_platforms


# TODO: also support non-target platforms and show their type in the list


class MyPlatformA(TargetPlatform):
    def __init__(self, features=None, config=None):
        super().__init__(
            "platform_a",
            features=features,
            config=config,
        )


class MyPlatformB(TargetPlatform):
    def __init__(self, features=None, config=None):
        super().__init__(
            "platform_b",
            features=features,
            config=config,
        )

    def get_supported_targets(self):
        return ["target_1"]

    def get_supported_backends(self):
        return ["backend_1"]


class MyPlatformC(TargetPlatform):
    def __init__(self, features=None, config=None):
        super().__init__(
            "platform_c",
            features=features,
            config=config,
        )

    def get_supported_targets(self):
        return ["target_1", "target_2"]

    def get_supported_backends(self):
        return ["backend_1", "backend_2"]


def test_platform_get_platforms_targets(fake_context):
    # empty
    with mock.patch("mlonmcu.platform.lookup.get_platform_names", return_value=[]):
        with mock.patch("mlonmcu.platform._platform.get_platforms", return_value={}):
            res = mlonmcu.platform.lookup.get_platforms_targets(fake_context, config={"a": "b"})
            assert len(res) == 0

    # single platform, no targets
    with mock.patch("mlonmcu.platform.lookup.get_platform_names", return_value=["platform_a"]):
        with mock.patch("mlonmcu.platform.lookup.get_platforms", return_value={"platform_a": MyPlatformA}):
            res = mlonmcu.platform.lookup.get_platforms_targets(fake_context, config={"a": "b"})
            assert len(res) == 1
            assert len(res["platform_a"]) == 0

    # single platform, single target
    with mock.patch("mlonmcu.platform.lookup.get_platform_names", return_value=["platform_b"]):
        with mock.patch("mlonmcu.platform.lookup.get_platforms", return_value={"platform_b": MyPlatformB}):
            res = mlonmcu.platform.lookup.get_platforms_targets(fake_context, config={"a": "b"})
            assert len(res) == 1
            assert len(res["platform_b"]) == 1

    # multiple platforms, overlapping targets
    with mock.patch("mlonmcu.platform.lookup.get_platform_names", return_value=["platform_b", "platform_c"]):
        with mock.patch(
            "mlonmcu.platform.lookup.get_platforms", return_value={"platform_b": MyPlatformB, "platform_c": MyPlatformC}
        ):
            res = mlonmcu.platform.lookup.get_platforms_targets(fake_context, config={"a": "b"})
            assert len(res) == 2
            assert len(res["platform_b"]) == 1
            assert len(res["platform_c"]) == 2


def test_platform_get_platforms_backends(fake_context):
    # empty
    with mock.patch("mlonmcu.platform.lookup.get_platform_names", return_value=[]):
        with mock.patch("mlonmcu.platform._platform.get_platforms", return_value={}):
            res = mlonmcu.platform.lookup.get_platforms_backends(fake_context, config={"a": "b"})
            assert len(res) == 0

    # single platform, no backends
    with mock.patch("mlonmcu.platform.lookup.get_platform_names", return_value=["platform_a"]):
        with mock.patch("mlonmcu.platform.lookup.get_platforms", return_value={"platform_a": MyPlatformA}):
            res = mlonmcu.platform.lookup.get_platforms_backends(fake_context, config={"a": "b"})
            assert len(res) == 1
            assert len(res["platform_a"]) == 0

    # single platform, single backend
    with mock.patch("mlonmcu.platform.lookup.get_platform_names", return_value=["platform_b"]):
        with mock.patch("mlonmcu.platform.lookup.get_platforms", return_value={"platform_b": MyPlatformB}):
            res = mlonmcu.platform.lookup.get_platforms_backends(fake_context, config={"a": "b"})
            assert len(res) == 1
            assert len(res["platform_b"]) == 1

    # multiple platforms, overlapping backends
    with mock.patch("mlonmcu.platform.lookup.get_platform_names", return_value=["platform_b", "platform_c"]):
        with mock.patch(
            "mlonmcu.platform.lookup.get_platforms", return_value={"platform_b": MyPlatformB, "platform_c": MyPlatformC}
        ):
            res = mlonmcu.platform.lookup.get_platforms_backends(fake_context, config={"a": "b"})
            assert len(res) == 2
            assert len(res["platform_b"]) == 1
            assert len(res["platform_c"]) == 2


def test_platform_print_summary(capsys, fake_context):
    # empty
    with mock.patch("mlonmcu.platform.lookup.get_platform_names", return_value=[]):
        with mock.patch("mlonmcu.platform._platform.get_platforms", return_value={}):
            mlonmcu.platform.lookup.print_summary(fake_context)
            out, err = capsys.readouterr()
            split = out.split("\n")
            assert "No platforms found" in split
            assert "No targets found" in split

    # single platform, no targets
    with mock.patch("mlonmcu.platform.lookup.get_platform_names", return_value=["platform_a"]):
        with mock.patch("mlonmcu.platform.lookup.get_platforms", return_value={"platform_a": MyPlatformA}):
            mlonmcu.platform.lookup.print_summary(fake_context)
            out, err = capsys.readouterr()
            split = out.split("\n")
            assert "  - platform_a" in split
            assert "No targets found" in split

    # single platform, single target
    with mock.patch("mlonmcu.platform.lookup.get_platform_names", return_value=["platform_b"]):
        with mock.patch("mlonmcu.platform.lookup.get_platforms", return_value={"platform_b": MyPlatformB}):
            mlonmcu.platform.lookup.print_summary(fake_context)
            out, err = capsys.readouterr()
            split = out.split("\n")
            assert "  - platform_b" in split
            assert "  - target_1 [platform_b]" in split
            assert "  - backend_1 [platform_b]" in split

    # multiple platforms, overlapping targets
    with mock.patch("mlonmcu.platform.lookup.get_platform_names", return_value=["platform_b", "platform_c"]):
        with mock.patch(
            "mlonmcu.platform.lookup.get_platforms", return_value={"platform_b": MyPlatformB, "platform_c": MyPlatformC}
        ):
            mlonmcu.platform.lookup.print_summary(fake_context)
            out, _ = capsys.readouterr()
            split = out.split("\n")
            assert "  - platform_b" in split
            assert "  - target_1 [platform_b, platform_c]" in split
            assert "  - target_2 [platform_c]" in split
            assert "  - backend_1 [platform_b, platform_c]" in split
            assert "  - backend_2 [platform_c]" in split


def test_platform_registry():
    before = len(get_platforms())
    register_platform("foo", MyPlatformA)
    register_platform("bar", MyPlatformB)
    assert len(get_platforms()) == before + 2
    with pytest.raises(RuntimeError):
        register_platform("foo", MyPlatformA)
    assert len(get_platforms()) == before + 2
    register_platform("foo", MyPlatformA, override=True)
    assert len(get_platforms()) == before + 2


def test_platform_get_platform_names(fake_context):
    def lookup_platform_configs(names_only=True):
        assert names_only
        return ["foo", "bar"]

    fake_context.environment.lookup_platform_configs = lookup_platform_configs
    res = mlonmcu.platform.lookup.get_platform_names(fake_context)
    assert res == ["foo", "bar"]



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest
import mock

from mlonmcu.target.common import execute, cli
from mlonmcu.target.target import Target
from mlonmcu.target import EtissPulpinoTarget, HostX86Target


class CustomTarget(Target):
    FEATURES = ["featureA"]

    def __init__(
        self,
        features=[],
        config={},
    ):
        super().__init__("custom", features=features, config=config)
        self.inspectProgram = "ls"
        self.inspectprogramArgs = []

    def exec(self, program, *args, **kwargs):
        return execute(program, *args, **kwargs)


# @pytest.mark.parametrize("ignore_output", [False, True])
# def test_target_common_execute(ignore_output, capsys):
#     # TODO: mock command line!
#     result = execute("/bin/date", ignore_output=ignore_output, live=False)
#     out, err = capsys.readouterr()
#     if ignore_output:
#         assert len(out) == 0
#     else:
#         assert len(out) > 0
#
# def test_target_common_cli_execute(capsys):
#     cli(target=CustomTarget, args=["exec", "/bin/date", "-f", "featureA"])
#     out, err = capsys.readouterr()
#
def _fake(*args, **kwargs):
    print("FAKE")


# @mock.patch('subprocess.run', side_effect=_fake)  # FIXME
@pytest.mark.parametrize("example_elf_file", ["elf-Linux-x64-bash"], indirect=True)
def test_target_common_cli_inspect(example_elf_file, capsys):
    cli(target=CustomTarget, args=["inspect", example_elf_file, "-c", "foo=bar"])
    out, err = capsys.readouterr()


@mock.patch("mlonmcu.target.common.execute", side_effect=_fake)  # FIXME
@pytest.mark.parametrize("example_elf_file", ["elf-Linux-x64-bash"], indirect=True)
def test_target_custom(mocked_execute, example_elf_file, capsys):
    t = CustomTarget()
    assert str(t) == "Target(custom)"

    # with pytest.raises(NotImplementedError):
    # t.exec("/bin/date")

    t.inspect(example_elf_file)
    # mocked_execute.assert_called_once_with("inspect", "program")


@mock.patch("mlonmcu.target.common.execute", side_effect=_fake)  # FIXME
@pytest.mark.parametrize("example_elf_file", ["elf-Linux-x64-bash"], indirect=True)
def test_target_base(mocked_execute, example_elf_file, capsys):
    t = Target("base")
    assert str(t) == "Target(base)"

    with pytest.raises(NotImplementedError):
        t.exec("/bin/date")

    t.inspect(example_elf_file)
    # mocked_execute.assert_called_once_with("inspect", "program")


def has_etiss_pulpino():
    return False


def has_riscv():
    return False


# TODO: needs etiss and riscv
@pytest.mark.skipif(not has_etiss_pulpino(), reason="requires etiss_pulpino")
@pytest.mark.skipif(not has_riscv(), reason="requires riscv")
@pytest.mark.parametrize("features", [[], ["etissdbg"], ["attach"], ["noattach"], ["trace"], ["v"]])
@pytest.mark.parametrize("example_elf_file", ["elf-Linux-ARM64-bash"], indirect=True)
def test_target_etiss_pulpino(features, example_elf_file, capsys):
    t = EtissPulpinoTarget(config={"riscv.dir": "foo", "etiss.dir": "bar"}, features=features)
    # t.exec("/bin/date")

    t.inspect(example_elf_file)


# TODO: needs gdbserver
def has_gdbserver():
    return False


def has_gdb():
    return False


@pytest.mark.skipif(not has_gdbserver(), reason="requires gdbserver")
@pytest.mark.skipif(not has_gdb(), reason="requires gdb")
@pytest.mark.parametrize("features", [[], ["attach"], ["noattach"]])
@pytest.mark.parametrize("example_elf_file", ["elf-Linux-x64-bash"], indirect=True)
def test_target_host_x86(features, example_elf_file, capsys):
    t = HostX86Target(features=features)
    t.exec("/bin/date")

    t.inspect(example_elf_file)



#!/usr/bin/env python3
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Modifications copyright (C) 2022  TUM Department of Electrical and Computer Engineering

"""MLonMCU Python requirements.txt generator.

This script generates a set of requirements.txt files (stored in `./requirements`) that describe
MLonMCU's Python dependencies.

## Pieces

MLonMCU can be roughly broken into these named pieces along the lines of Python dependencies:

- "core": A core piece, which is intended to be buildable with very few external dependencies. Users
  can use Relay, compile models, and run autotuning with this part.
- Extra features (i.e. TVM). These enhance MLonMCU's functionality, but aren't required
  for basic operation.

## What this tool does

From these pieces, this tool builds:
 - requirements/<name>.txt - Python dependencies for each named piece above, `<name>` is the same as
   the quoted piece name.
 - requirements/all.txt - Consolidated Python dependencies for all pieces, excluding dev below.
 - requirements/dev.txt - Python dependencies needed to develop MLONMCU, such as lint and test tools.

The data representing each piece is contained in the two maps below.
"""

import argparse
import collections
import os
import re
import textwrap
import sys
import typing


RequirementsByPieceType = typing.List[typing.Tuple[str, typing.Tuple[str, typing.List[str]]]]


# Maps named MLonMCU piece (see description above) to a list of names of Python packages. Please use
# alphabetical order for each package list, and do not add version constraints here!
REQUIREMENTS_BY_PIECE: RequirementsByPieceType = [
    # Base requirements needed to install MLonMCU.
    (
        "core",
        (
            "Base requirements needed to install MLonMCU",
            [
                "GitPython",
                "Pillow",  # convert data
                "filelock",
                "humanize",
                "jinja2",
                "networkx",
                "numpy",
                "openpyxl",  # xlsx reports
                "pandas",
                "pyelftools",
                "pyyaml",
                "tqdm",
                "xdg",
                "xlsxwriter",  # xlsx reports
                "xlwt",  # xlsx reports
            ],
        ),
    ),
    # Provide support for espidf.
    (
        "espidf",
        (
            "Requirements for using espidf",
            [
                "bitstring",
                "click",
                "construct",
                "cryptography",
                "ecdsa",
                "future",
                "gdbgui",
                "idf-component-manager",
                "itsdangerous",  # there are two espidf in origin requirements.txt this is from the 2nd
                "jinja2",  # there are two espidf in origin requirements.txt this is from the 2nd
                "kconfiglib",
                "psutil",
                "pygdbmi",
                "pyparsing",
                "pyserial",  # for custom monitor
                "python-socketio",
                "reedsolo",
            ],
        ),
    ),
    # Provide support for etiss.
    (
        "etiss",
        (
            "Requirements for using etiss",
            ["humanize", "pyelftools"],
        ),
    ),
    ("gvsoc_pulp", ("Requirements for gvsoc", ["PrettyTable"])),
    # Provide support for microtvm.
    (
        "microtvm",
        (
            "Requirements for using microtvm",
            ["matplotlib", "pyserial", "pyusb"],
        ),
    ),
    # Provide support for moiopt.
    ("moiopt", ("Requirements for moiopt", ["ortools"])),
    # Provide support for onnx.
    ("onnx", ("Requirements for onnx", ["onnx"])),
    # Provide support for relay visualization.
    ("relay-visualization", ("Requirements for relay visualization", ["relayviz"])),
    # Provide support for tflite.
    (
        "tflite",
        (
            "Requirements for using tflite",
            [
                "tflite",
            ],
        ),
    ),
    # Provide support for tflm.
    (
        "tflm",
        (
            "Requirements for using tflm",
            [
                "Pillow",
            ],
        ),
    ),
    # Provide support for tvm.
    (
        "tvm",
        (
            "Requirements for using tvm",
            [
                "attrs",
                "cloudpickle",
                "decorator",
                "graphviz",  # for relay visualization
                "matplotlib",
                "numpy",
                "psutil",
                "scipy",
                "synr",
                "tornado",
                "typing_extensions",
            ],
        ),
    ),
    ("tvm-autotuning", ("Requirements for using tvm autotuning", ["xgboost"])),  # for autotuning on some target
    # Provide support for visualize.
    (
        "visualize",
        (
            "Requirements for using visualize",
            ["matplotlib"],
        ),
    ),
    ("visualize_tflite", ("Requirements for visualization", ["tensorflow"])),
    # Provide support for zephyr.
    (
        "zephyr",
        (
            "Requirements for using zephyr",
            [
                "pyserial",  # for custom monitor
            ],
        ),
    ),
]

ConstraintsType = typing.List[typing.Tuple[str, typing.Union[None, str]]]

# Maps a named Python package (which should appear in REQUIREMENTS_BY_PIECE above) to a
# semver or pip version constraint. Semver constraints are translated into requirements.txt-friendly
# constraints.
#
# These constraints serve only to record technical reasons why a particular version can't be used.
# They are the default install_requires used in setup.py. These can be further narrowed to restrict
# dependencies to those tested or used in CI; however, that process is not done here.
#
# Policy for constraints listed here:
# 1. Each package specified in REQUIREMENTS_BY_PIECE must be included here.
# 2. If MLonMCU will functionally break against an old version of a dependency, specify a >= relation
#    here. Include a comment linking to context or explaining why the constraint is in place.
CONSTRAINTS = [
    ("GitPython", None),
    ("Pillow", None),
    ("bitstring", ">=3.1.6"),
    ("click", ">=7.0"),
    ("cloudpickle", None),
    ("construct", "==2.10.54"),  # From PR #213.
    ("cryptography", ">=2.1.4"),
    ("decorator", None),
    ("ecdsa", ">=0.16.0"),
    ("filelock", None),
    ("future", ">=0.15.2"),
    ("gdbgui", "==0.13.2.0"),
    ("graphviz", None),
    ("humanize", None),
    ("idf-component-manager", "~=1.0"),
    ("itsdangerous", "<2.1"),
    ("jinja2", ">=3.1.3"),
    ("kconfiglib", "==13.7.1"),
    ("matplotlib", None),
    ("networkx", None),
    ("numpy", "<2.0"),
    ("onnx", None),
    ("openpyxl", None),
    ("ortools", None),
    ("pandas", None),
    ("prettytable", None),
    ("psutil", None),
    ("pyelftools", None),
    ("pygdbmi", "<=0.9.0.2"),
    ("pyparsing", ">=2.0.3,<2.4.0"),
    ("pyserial", None),
    ("python-socketio", "<5"),
    ("pyusb", None),
    ("pyyaml", None),
    ("reedsolo", ">=1.5.3,<=1.5.4"),
    ("scipy", None),
    ("synr", None),
    ("tensorflow", ">=2.11.1"),
    ("tflite", ">=2.4.0"),
    ("tornado", None),
    ("typing_extensions", None),
    ("xdg", None),
    ("xgboost", ">=1.7.0"),
    ("xlsxwriter", None),
    ("xlwt", None),
]

################################################################################
# End of configuration options.
################################################################################


# Required keys in REQUIREMENTS_BY_PIECE.
REQUIRED_PIECES: typing.List[str] = ["core", "dev"]

# Regex to validates piece names.
PIECE_REGEX: typing.Pattern = re.compile(r"^[a-z0-9][a-z0-9-]*", re.IGNORECASE)

# Regex to match a constraint specification. Multiple constraints are not supported.
CONSTRAINT_REGEX: typing.Pattern = re.compile(r"(?:\^|\<|(?:~=)|(?:<=)|(?:==)|(?:>=)|\>)[^<>=\^,]+")

# Regex for parsing semantic versions. See
# https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
SEMVER_REGEX: typing.Pattern = re.compile(
    r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)"
    r"(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))"
    r"?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
)


def validate_requirements_by_piece() -> typing.List[str]:
    """Validate REQUIREMENTS_BY_PIECE, returning a list of problems.

    Returns
    -------
    list[str] :
        A list of strings, each one describing a distinct problem with REQUIREMENTS_BY_PIECE.
    """
    problems = []

    unseen_required_pieces = set(REQUIRED_PIECES)
    seen_pieces = set()

    # Ensure that core is listed first and dev is listed last.
    saw_core = False
    saw_dev = False

    if not isinstance(REQUIREMENTS_BY_PIECE, (list, tuple)):
        problems.append(f"must be list or tuple, see {REQUIREMENTS_BY_PIECE!r}")
        return problems

    for piece, value in REQUIREMENTS_BY_PIECE:
        if not isinstance(piece, str):
            problems.append(f"piece {piece!r}: must be str")
            continue

        if piece in unseen_required_pieces:
            unseen_required_pieces.remove(piece)

        piece_lower = piece.lower()
        if piece_lower in seen_pieces:
            problems.append(f"piece {piece}: listed twice")

        seen_pieces.add(piece_lower)

        if not saw_core and piece != "core":
            problems.append(f'piece {piece}: must list after "core" (core must be first)')
        elif piece == "core":
            saw_core = True

        if saw_dev:
            problems.append(f'piece {piece}: must list before "dev" (dev must be last)')
        elif piece == "dev":
            saw_dev = True

        if not isinstance(value, (tuple, list)) or len(value) != 2:
            problems.append(
                f'piece {piece}: should be formatted like ("{piece}", ("<requirements.txt comment>", ["dep1", "dep2", '
                f"...])). got: {value!r} "
            )
            continue

        description, deps = value

        if not isinstance(description, str):
            problems.append(f"piece {piece}: description should be a string, got {description!r}")

        if not isinstance(deps, (list, tuple)) or any(not isinstance(d, str) for d in deps):
            problems.append(f"piece {piece}: deps should be a list of strings, got {deps!r}")
            continue

        if list(sorted(deps)) != list(deps):
            problems.append(f"piece {piece}: deps must be sorted. Correct order:\n  {list(sorted(deps))!r}")

        piece_deps = set()
        for d in deps:
            if CONSTRAINT_REGEX.search(d):
                problems.append(
                    f"piece {piece}: dependency {d} should not specify a version. Add it to CONSTRAINTS instead."
                )

            if d.lower() in piece_deps:
                problems.append(f"piece {piece}: dependency {d} listed twice")

            piece_deps.add(d.lower())

    extras_pieces = [k for (k, _) in REQUIREMENTS_BY_PIECE if k not in ("dev", "core") if isinstance(k, str)]
    sorted_extras_pieces = list(sorted(extras_pieces))
    if sorted_extras_pieces != list(extras_pieces):
        problems.append(f'pieces other than "core" and "dev" must appear in alphabetical order: {sorted_extras_pieces}')

    return problems


def parse_semver(package: str, constraint: str, problems: typing.List[str]) -> typing.Tuple[typing.List[str], int, int]:
    """Parse a semantic versioning constraint of the form "^X.[.Y[.Z[...]]]]"

    Parameters
    ----------
    package : str
        Name of the package specifying this constraint, for reporting problems.
    constraint : str
        The semver constraint. Must start with "^"
    problems : List[str]
        A list of strings describing problems that have occurred validating the configuration.
        Problems encountered while validating constraint are appended to this list.

    Returns
    -------
    tuple[list[str], int, int] :
        A 3-tuple. The first element is a list containing an entry for each component in the
        semver string (components separated by "."). The second element is the index of the
        component in the list which must not change to meet the semver constraint. The third element
        is an integer, the numeric value of the changing component (this can be non-trivial when
        the patch is the changing part but pre-, post-release, or build metadta.

        See "Caret requirements" at https://python-poetry.org/docs/versions/.
    """
    m = SEMVER_REGEX.match(constraint[1:])
    if not m:
        problems.append(f"{package}: invalid semver constraint {constraint}")
        return [], 0, 0

    min_ver_parts = [
        m.group("major"),
        m.group("minor"),
        m.group("patch")
        + (f"-{m.group('prerelease')}" if m.group("prerelease") else "")
        + (f"+{m.group('buildmetadata')}" if m.group("buildmetadata") else ""),
    ]

    # Major/minor version handling is simple
    for i, p in enumerate(min_ver_parts[:2]):
        x = int(p.strip())
        if x:
            return min_ver_parts, i, x

    # For patch version, consult only the numeric patch
    if m.group("patch"):
        patch_int = int(m.group("patch"))
        if patch_int or min_ver_parts[2] != m.group("patch"):
            return min_ver_parts, 2, patch_int

    # All 0's
    return min_ver_parts, 0, 0


def validate_constraints() -> typing.List[str]:
    """Validate CONSTRAINTS, returning a list of problems found.

    Returns
    -------
    list[str] :
        A list of strings, each one describing a distinct problem found in CONSTRAINTS.
    """
    problems = []

    if not isinstance(CONSTRAINTS, (list, tuple)):
        problems.append(f"must be list or tuple, see: {CONSTRAINTS!r}")

    seen_packages = set()
    all_deps = set()
    for _, (_, deps) in REQUIREMENTS_BY_PIECE:
        for d in deps:
            all_deps.add(d.lower())

    for package, constraint in CONSTRAINTS:
        if package in seen_packages:
            problems.append(f"{package}: specified twice")
        seen_packages.add(package)

        if package.lower() not in all_deps:
            problems.append(f"{package}: not specified in REQUIREMENTS_BY_PIECE")

        if constraint is None:  # None is just a placeholder that allows for comments.
            continue

        if not CONSTRAINT_REGEX.match(constraint):
            problems.append(f'{package}: constraint "{constraint}" does not look like a valid constraint')

        if constraint.startswith("^"):
            parse_semver(package, constraint, problems)

    all_constrained_packages = [p for (p, _) in CONSTRAINTS]
    sorted_constrained_packages = list(sorted(all_constrained_packages))
    if sorted_constrained_packages != all_constrained_packages:
        problems.append(f"CONSTRAINTS entries should be in this sorted order: {sorted_constrained_packages}")

    return problems


class ValidationError(Exception):
    """Raised when a validation error occurs."""

    @staticmethod
    def format_problems(config: str, problems: typing.List[str]) -> str:
        """Format a list of problems with a global config variable into human-readable output.

        Parameters
        ----------
        config : str
            Name of the global configuration variable of concern. Prepended to the output.
        problems: list[str]
            A list of strings, each one a distinct problem with that config variable.

        Returns
        -------
        str :
            A human-readable string suitable for console, listing the problems as bullet points.
        """
        formatted = []
        for p in problems:
            assert isinstance(p, str), f"problems element not a str: {p}"
            formatted.append(
                "\n".join(textwrap.wrap(f"{config}: {p}", width=80, initial_indent=" * ", subsequent_indent="   "))
            )

        return "\n".join(formatted)

    def __init__(self, config: str, problems: typing.List[str]):
        """Describes an error that occurs validating one of the global config variables.

        Parameters
        ----------
        config : str
            Name of the global configuration variable of concern. Prepended to the output.
        problems: list[str]
            A list of strings, each one a distinct problem with that config variable.
        """
        super(ValidationError, self).__init__(self.format_problems(config, problems))
        self.problems = problems


def validate_or_raise():
    problems = validate_requirements_by_piece()
    if problems:
        raise ValidationError("REQUIREMENTS_BY_PIECE", problems)

    problems = validate_constraints()
    if problems:
        raise ValidationError("CONSTRAINTS", problems)


def semver_to_requirements(dep: str, constraint: str, joined_deps: typing.List[str]):
    """Convert a SemVer-style constraint to a setuptools-compatible constraint.

    Parameters
    ----------
    dep : str
        Name of the PyPI package to depend on.
    constraint : str
        The SemVer constraint, of the form "^<semver constraint>"
    joined_deps : list[str]
        A list of strings, each a setuptools-compatible constraint which could be written to
        a line in requirements.txt. The converted constraint is appended to this list.
    """
    problems: typing.List[str] = []
    min_ver_parts, fixed_index, fixed_part = parse_semver(dep, constraint, problems)
    text_problems = "\n" + "\n".join(f" * {p}" for p in problems)
    assert not problems, f"should not happen: validated semver {constraint} parses with problems:{text_problems}"

    max_ver_parts = (
        min_ver_parts[:fixed_index] + [str(fixed_part + 1)] + ["0" for _ in min_ver_parts[fixed_index + 1 :]]
    )
    joined_deps.append(f'{dep}>={".".join(min_ver_parts)},<{".".join(max_ver_parts)}')


def join_requirements() -> typing.Dict[str, typing.Tuple[str, typing.List[str]]]:
    """Validate, then join REQUIRMENTS_BY_PIECE against CONSTRAINTS and return the result.

    Returns
    -------
    An OrderedDict containing REQUIREMENTS_BY_PIECE, except any dependency mentioned in CONSTRAINTS
    is replaced by a setuptools-compatible constraint.
    """
    validate_or_raise()

    constraints_map = collections.OrderedDict([(p.lower(), c) for (p, c) in CONSTRAINTS])

    to_return = collections.OrderedDict()
    all_deps = set()
    for piece, (description, deps) in REQUIREMENTS_BY_PIECE:
        joined_deps = []
        for d in deps:
            constraint = constraints_map.get(d.lower())
            if constraint is None:
                joined_deps.append(d)
                continue

            if constraint[0] == "^":
                semver_to_requirements(d, constraint, joined_deps)
            else:
                joined_deps.append(f"{d}{constraint}")

        if piece != "dev":
            all_deps.update(joined_deps)

        to_return[piece] = (description, joined_deps)

    to_return["all-prod"] = (
        "Combined dependencies for all MLONMCU pieces, excluding dev",
        list(sorted(all_deps)),
    )

    return to_return


def join_and_write_requirements(args: argparse.Namespace):
    try:
        joined_deps = join_requirements()
    except ValidationError as e:
        print(f"ERROR: invalid requirements configuration in {__file__}:", file=sys.stderr)
        print(str(e), file=sys.stderr)
        sys.exit(2)

    if args.lint:
        sys.exit(0)

    output_dir = os.path.join(os.path.dirname(__file__), "requirements")
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    elif not os.path.isdir(output_dir):
        print(
            f"ERROR: output directory {output_dir} exists but is not a dir. Delete it",
            file=sys.stderr,
        )
        sys.exit(2)

    for piece, (description, deps) in joined_deps.items():
        with open(os.path.join(output_dir, f"{piece}.txt"), "w") as f:
            f.write(
                f"# AUTOGENERATED by python/gen_requirements.py{os.linesep}#{os.linesep}# {description}{os.linesep}"
            )
            for d in deps:
                f.write(f"{d}{os.linesep}")


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--lint", action="store_true", help="Just lint dependencies, don't generate anything")
    return parser.parse_args()


def main():
    args = parse_args()
    join_and_write_requirements(args)


if __name__ == "__main__":
    main()



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_tensorflow(context: MlonMcuContext, params=None):
    return context.environment.has_framework("tflm")


@Tasks.provides(["tf.src_dir"])
@Tasks.validate(_validate_tensorflow)
@Tasks.register(category=TaskType.FRAMEWORK)
def clone_tensorflow(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the TF/TFLM repository."""
    tfName = utils.makeDirName("tf")
    tfSrcDir = context.environment.paths["deps"].path / "src" / tfName
    if rebuild or not utils.is_populated(tfSrcDir):
        tfRepo = context.environment.repos["tensorflow"]
        utils.clone_wrapper(tfRepo, tfSrcDir, refresh=rebuild)
    context.cache["tf.src_dir"] = tfSrcDir


@Tasks.needs(["tf.src_dir"])
@Tasks.provides(["tf.dl_dir", "tf.lib_path"])
# @Tasks.param("dbg", False)
@Tasks.param("dbg", True)
@Tasks.validate(_validate_tensorflow)
@Tasks.register(category=TaskType.FRAMEWORK)
def build_tensorflow(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Download tensorflow dependencies and build lib."""
    if not params:
        params = {}
    flags = utils.makeFlags((params["dbg"], "dbg"))
    # tfName = utils.makeDirName("tf", flags=flags)
    tfSrcDir = context.cache["tf.src_dir"]
    tflmDir = Path(tfSrcDir) / "tensorflow" / "lite" / "micro"
    tflmBuildDir = tflmDir / "tools" / "make"
    tflmDownloadsDir = tflmBuildDir / "downloads"
    if params["dbg"]:
        tflmLib = (
            tflmBuildDir / "gen" / "linux_x86_64" / "lib" / "libtensorflow-microlite.a"
        )  # FIXME: add _dbg suffix is possible
    else:
        tflmLib = tflmBuildDir / "gen" / "linux_x86_64" / "lib" / "libtensorflow-microlite.a"
    # if rebuild or not tflmLib.is_file() or not utils.is_populated(tflmDownloadsDir):
    if rebuild and utils.is_populated(tflmDownloadsDir):
        utils.make(
            "-f",
            str(tflmDir / "tools" / "make" / "Makefile"),
            "clean_downloads",
            cwd=tfSrcDir,
            live=verbose,
        )
    if rebuild or not utils.is_populated(tflmDownloadsDir):
        tfDbgArg = ["BUILD_TYPE=debug"] if params["dbg"] else []
        utils.make(
            "-f",
            str(tflmDir / "tools" / "make" / "Makefile"),
            "third_party_downloads",
            threads=threads,
            *tfDbgArg,
            cwd=tfSrcDir,
            live=verbose,
        )
    context.cache["tf.dl_dir"] = tflmDownloadsDir
    context.cache["tf.lib_path", flags] = tflmLib  # ignore!


def _validate_tflite_pack(context: MlonMcuContext, params=None):
    return context.environment.has_frontend("tflite") and context.environment.has_feature("split_layers")


@Tasks.provides(["tflite_pack.exe", "tflite_pack.src_dir"])
@Tasks.validate(_validate_tflite_pack)
@Tasks.register(category=TaskType.FEATURE)
def clone_tflite_pack(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the tflite packing utilities."""
    name = utils.makeDirName("tflite_pack")
    srcDir = context.environment.paths["deps"].path / "src" / name
    if rebuild or not utils.is_populated(srcDir):
        repo = context.environment.repos["tflite_pack"]
        utils.clone_wrapper(repo, srcDir, refresh=rebuild)
    context.cache["tflite_pack.src_dir"] = srcDir
    context.cache["tflite_pack.exe"] = srcDir / "run.sh"


@Tasks.needs(["tflite_pack.src_dir"])
@Tasks.provides(["tflite_pack.exe"])
@Tasks.validate(_validate_tflite_pack)
@Tasks.register(category=TaskType.FEATURE)
def install_tflite_pack(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Install the tflite packing utilities."""
    name = utils.makeDirName("tflite_pack")
    srcDir = Path(context.cache["tflite_pack.src_dir"])
    installDir = context.environment.paths["deps"].path / "install" / name
    if rebuild or not utils.is_populated(installDir):
        installScript = srcDir / "install.sh"
        utils.exec_getout(installScript, installDir, live=verbose)
    context.cache["tflite_pack.exe"] = installDir / "run.sh"
